@@ -159,11 +159,53 @@ static void add_node(nx_cfg_parser_ctx_t *ctx, nx_directive_t *newdir)
 
 
 
+/*
+ * This function returns TRUE if the keyword is a special block
+ * that should not be parsed so that we can have anything (e.g. XML)
+ * inside. This is a ugly to hardcode keywords here but module dso is not loaded
+ * at this stage so we can't put this there.
+ */
+
+static boolean need_body_parsed(boolean current_state, const char *tag)
+{
+    typedef struct parse_keywords
+    {
+	const char *name;
+	boolean state;
+    } parse_keywords;
+
+    static const parse_keywords keywords[] =
+    {
+	{ "<Exec>", TRUE },
+	{ "</Exec>", FALSE },
+	{ "<QueryXML>", TRUE }, // this is for im_msvistalog
+	{ "</QueryXML>", FALSE }, // this is for im_msvistalog
+	{ NULL, FALSE }
+    };
+    int i;
+
+    if ( (current_state == FALSE) && (tag[0] != '<') )
+    {
+	return ( FALSE );
+    }
+
+    for ( i = 0; keywords[i].name != NULL; i++ )
+    {
+	if ( (keywords[i].state != current_state) && (strcasecmp(tag, keywords[i].name) == 0) )
+	{
+	    return ( keywords[i].state );
+	}
+    }
+
+    return ( current_state );
+}
+
+
 
 /*
  * return the first element parsed
  */
-static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx)
+static nx_directive_t *nx_cfg_parse_line(const char *l, int trimmedspace, nx_cfg_parser_ctx_t *ctx)
 {
     const char *args;
     char *cmd_name;
@@ -172,10 +214,12 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
     nx_cfg_parser_ctx_t cfg_ctx;
     nx_directive_t *retval = NULL;
     nx_exception_t e;
+    boolean dont_parse = FALSE;
+    boolean prev_dont_parse;
 
     ASSERT(ctx != NULL);
 
-    if ( (*l == '#') || (*l == '\0') )
+    if ( *l == '\0' )
     {
         return ( NULL );
     }
@@ -183,7 +227,27 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
     args = l;
 
     cmd_name = nx_cfg_getword(ctx->pool, &args);
-    ASSERT(*cmd_name != '\0');
+    if ( *cmd_name == '\0' )
+    {
+	return ( NULL );
+    }
+
+    prev_dont_parse = ctx->dont_parse;
+    ctx->dont_parse = need_body_parsed(ctx->dont_parse, cmd_name);
+
+    if ( (prev_dont_parse == TRUE) && (ctx->dont_parse == TRUE) )
+    { // append 
+	ASSERT(ctx->current != NULL);
+	ASSERT(ctx->current->args != NULL);
+	ctx->current->args = apr_pstrcat(ctx->pool, ctx->current->args, l, NULL);
+	
+	return ( NULL );
+    }
+
+    if ( *l == '#' )
+    { // ignore comment lines 
+        return ( NULL );
+    }
 
     if ( strcasecmp(cmd_name, "include") == 0 )
     {
@@ -262,7 +326,8 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
 	}
 	else
 	{
-	    incfile.name = args;
+	    incfile.name = apr_pstrdup(ctx->pool, args);
+
 	    try
 	    {
 	        nx_cfg_open_file(&incfile, ctx->pool);
@@ -273,6 +338,7 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
 			    ctx->cfg->name, ctx->cfg->line_num);
 	    }
 	    retval = nx_cfg_parse(&cfg_ctx);
+
 	    ctx->current = cfg_ctx.current;
 	    ctx->next_is_child = cfg_ctx.next_is_child;
 	    ctx->defines = cfg_ctx.defines;
@@ -280,6 +346,11 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
 	return ( retval );
     }
 
+    newdir = apr_pcalloc(ctx->pool, sizeof(nx_directive_t));
+    ASSERT(args > l);
+	   
+    newdir->argsstart = (int) (args - l) + trimmedspace;
+
     if ( cmd_name[1] != '/' )
     {
         char *lastc = cmd_name + strlen(cmd_name) - 1;
@@ -293,13 +364,18 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
         }
     }
 
-    newdir = apr_pcalloc(ctx->pool, sizeof(nx_directive_t));
     newdir->filename = ctx->cfg->name;
     newdir->line_num = ctx->cfg->line_num;
     newdir->directive = apr_pstrdup(ctx->pool, cmd_name);
-    newdir->args = apr_pstrdup(ctx->pool, args);
-    newdir->argsstart = (int) (args - l);
-    
+    if ( ctx->dont_parse == TRUE )
+    {
+	newdir->args = apr_pstrcat(ctx->pool, "\n", args, NULL);
+    }
+    else
+    {
+	newdir->args = apr_pstrdup(ctx->pool, args);
+    }
+
     if ( strcasecmp(cmd_name, "define") == 0 )
     {
 	newdir->directive = nx_cfg_getword(ctx->pool, (const char **) &(newdir->args));
@@ -338,7 +414,7 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
 	    return ( newdir );
         }
         else
-	{ // closing tag
+	{ // closing tag </
 	    char *bracket = cmd_name + strlen(cmd_name) - 1;
 
 	    if ( ctx->level <= 0 )
@@ -363,7 +439,10 @@ static nx_directive_t *nx_cfg_parse_line(const char *l, nx_cfg_parser_ctx_t *ctx
 	    }
 
 	    /* done with this section; move up a level */
-	    ctx->current = ctx->current->parent;
+	    if ( prev_dont_parse == FALSE )
+	    {
+		ctx->current = ctx->current->parent;
+	    }
 	    ASSERT(ctx->current != NULL);
 	    for ( ; ctx->current->next != NULL; ctx->current = ctx->current->next );
 	    (ctx->level)--;
@@ -421,174 +500,107 @@ static apr_status_t cfg_close(void *param)
 
 
 
-/* Read one line from open ap_configfile_t, strip LF, increase line number */
+/* Read one line from opened nx_configfile_t, strip LF, increase line number */
 /* If custom handler does not define a getstr() function, read char by char */
 
-int nx_cfg_getline(char *buf, int bufsize, nx_cfgfile_t *cfp)
+static int nx_cfg_getline(char *buf,
+			  int *trimmedspace,
+			  int bufsize,
+			  nx_cfgfile_t *cfp,
+			  boolean dont_parse)
 {
+    char *src, *dst;
+    char *cp;
+    char *cbuf = buf;
+    int cbufsize = bufsize;
+
     ASSERT(buf != NULL);
-
-    /* If a "get string" function is defined, use it */
-    if ( cfp->getstr != NULL )
+    ASSERT(cfp->getstr != NULL);
+    ASSERT(trimmedspace != NULL);
+    
+    *trimmedspace = 0;
+    for ( ; ; )
     {
-        char *src, *dst;
-        char *cp;
-        char *cbuf = buf;
-        int cbufsize = bufsize;
-
-        while ( 1 )
+	(cfp->line_num)++;
+	if ( cfp->getstr(cbuf, cbufsize, cfp->param) == NULL )
 	{
-            ++cfp->line_num;
-            if ( cfp->getstr(cbuf, cbufsize, cfp->param) == NULL )
+	    return 1;
+	}
+	/*
+	 *  check for line continuation,
+	 *  i.e. match [^\\]\\[\r]\n only
+	 */
+	if ( dont_parse == FALSE )
+	{
+	    cp = cbuf;
+	    while ( (cp < cbuf+cbufsize) && (*cp != '\0') )
 	    {
-                return 1;
+		cp++;
 	    }
-            /*
-             *  check for line continuation,
-             *  i.e. match [^\\]\\[\r]\n only
-             */
-            cp = cbuf;
-            while ( (cp < cbuf+cbufsize) && (*cp != '\0') )
+	    if ( (cp > cbuf) && (cp[-1] == APR_ASCII_LF) )
 	    {
-                cp++;
-	    }
-            if ( (cp > cbuf) && (cp[-1] == APR_ASCII_LF) )
-	    {
-                cp--;
-                if ( (cp > cbuf) && (cp[-1] == APR_ASCII_CR) )
+		cp--;
+		if ( (cp > cbuf) && (cp[-1] == APR_ASCII_CR) )
 		{
-                    cp--;
+		    cp--;
 		}
-                if ( (cp > cbuf) && (cp[-1] == '\\') )
+		if ( (cp > cbuf) && (cp[-1] == '\\') )
 		{
-                    cp--;
-                    if (! ((cp > cbuf) && (cp[-1] == '\\')) )
+		    cp--;
+		    if (! ((cp > cbuf) && (cp[-1] == '\\')) )
 		    {
-                        /*
-                         * line continuation requested -
-                         * then remove backslash and continue
-                         */
-                        cbufsize -= (int) (cp - cbuf);
-                        cbuf = cp;
-                        continue;
-                    }
-                    else
+			/*
+			 * line continuation requested -
+			 * then remove backslash and continue
+			 */
+			cbufsize -= (int) (cp - cbuf);
+			cbuf = cp;
+			continue;
+		    }
+		    else
 		    {
-                        /*
-                         * no real continuation because escaped -
-                         * then just remove escape character
-                         */
-                        for ( ; (cp < cbuf + cbufsize) && (*cp != '\0'); cp++ )
+			/*
+			 * no real continuation because escaped -
+			 * then just remove escape character
+			 */
+			for ( ; (cp < cbuf + cbufsize) && (*cp != '\0'); cp++ )
 			{
-                            cp[0] = cp[1];
+			    cp[0] = cp[1];
 			}
-                    }
-                }
-            }
-            break;
-        }
-
-        /*
-         * Leading and trailing white space is eliminated completely
-         */
-        src = buf;
-        while ( apr_isspace(*src) )
-	{
-            src++;
+		    }
+		}
+	    }
 	}
-        /* blast trailing whitespace */
-        dst = &src[strlen(src)];
-        while ( (--dst >= src) && apr_isspace(*dst) )
-	{
-            *dst = '\0';
-	}
-        /* Zap leading whitespace by shifting */
-        if ( src != buf )
-	{
-            for ( dst = buf; (*dst++ = *src++) != '\0'; );
-	}
-        //log_debug("Read config: %s", buf);
-        return 0;
+	break;
     }
-    else
+
+    /*
+     * Leading and trailing white space is eliminated completely
+     */
+    src = buf;
+    if ( dont_parse == FALSE )
     {
-        /* No "get string" function defined; read character by character */
-        int c;
-        int i = 0;
-
-        buf[0] = '\0';
-        /* skip leading whitespace */
-        do
+	/* Zap leading whitespace by shifting */
+	while ( apr_isspace(*src) )
 	{
-            c = cfp->getch(cfp->param);
-        }
-	while ( (c == '\t') || (c == ' ') );
-
-        if ( c == EOF )
-	{
-            return 1;
+	    src++;
+	    (*trimmedspace)++;
 	}
 
-        if ( bufsize < 2 )
+	/* blast trailing whitespace */
+	dst = &src[strlen(src)];
+	while ( (--dst >= src) && apr_isspace(*dst) )
 	{
-            /* too small, assume caller is crazy */
-            return 1;
-        }
-
-        while ( 1 )
-	{
-            if ( (c == '\t') || (c == ' ') )
-	    {
-                buf[i++] = ' ';
-                while ( (c == '\t') || (c == ' ') )
-		{
-                    c = cfp->getch(cfp->param);
-		}
-            }
-            if ( c == APR_ASCII_CR )
-	    {
-                /* silently ignore CR (_assume_ that a LF follows) */
-                c = cfp->getch(cfp->param);
-            }
-            if ( c == APR_ASCII_LF )
-	    {
-                /* increase line number and return on LF */
-                ++cfp->line_num;
-            }
-            if ( (c == EOF) || (c == 0x4) || (c == APR_ASCII_LF) || (i >= (bufsize - 2)) )
-	    {
-                /*
-                 *  check for line continuation
-                 */
-                if ( (i > 0) && (buf[i-1] == '\\') )
-		{
-                    i--;
-                    if ( !((i > 0) && (buf[i-1] == '\\')) )
-		    {
-                        /* line is continued */
-                        c = cfp->getch(cfp->param);
-                        continue;
-                    }
-                    /* else nothing needs be done because
-                     * then the backslash is escaped and
-                     * we just strip to a single one
-                     */
-                }
-                /* blast trailing whitespace */
-                while ( (i > 0) && apr_isspace(buf[i - 1]) )
-		{ 
-                    i--;
-		}
-                buf[i] = '\0';
-
-                //log_debug("Read config: %s", buf);
-                return 0;
-            }
-            buf[i] = (char) c;
-            ++i;
-            c = cfp->getch(cfp->param);
-        }
+	    *dst = '\0';
+	}
     }
+
+    if ( src != buf )
+    {
+	for ( dst = buf; (*dst++ = *src++) != '\0'; );
+    }
+    //log_debug("Read config: %s", buf);
+    return 0;
 }
 
 
@@ -662,17 +674,18 @@ nx_directive_t *nx_cfg_parse(nx_cfg_parser_ctx_t *ctx)
     nx_directive_t * volatile curr = NULL;
     nx_directive_t * volatile retval = NULL;
     nx_exception_t e;
+    int trimmedspace = 0;
 
     ASSERT(ctx != NULL);
 
     l = apr_palloc(ctx->pool, MAX_STRING_LEN);
 
-    while ( ! (nx_cfg_getline(l, MAX_STRING_LEN, ctx->cfg)) )
+    while ( ! (nx_cfg_getline(l, &trimmedspace, MAX_STRING_LEN, ctx->cfg, ctx->dont_parse)) )
     {
 	try
 	{
 	    nx_cfg_subst_defines(l, MAX_STRING_LEN, ctx->defines);
-	    curr = nx_cfg_parse_line(l, ctx);
+	    curr = nx_cfg_parse_line(l, trimmedspace, ctx);
 	}
 	catch(e)
 	{
@@ -684,7 +697,11 @@ nx_directive_t *nx_cfg_parse(nx_cfg_parser_ctx_t *ctx)
 	    retval = curr;
 	}
     }
-    
+    if ( ctx->level != 0 )
+    {
+	throw_msg("Expected </%s> at %s:%d", (ctx->levels[ctx->level - 1])->directive,
+		  ctx->cfg->name, ctx->cfg->line_num);
+    }
     ctx->cfg->close(ctx->cfg->param);
 
     return ( retval );
@@ -711,8 +728,9 @@ void nx_cfg_dump(const nx_directive_t *conftree, int level)
     {
 	apr_snprintf(buf + i, sizeof(buf) - (unsigned int)i, "%s %s", curr->directive,
 		     curr->args == NULL ? "" : curr->args);
+
 	printf("%s\n", buf);
-	
+
 	if ( curr->first_child != NULL )
 	{
 	    nx_cfg_dump(curr->first_child, level + 1);
