@@ -6,12 +6,14 @@
  */
 
 #include <apr_lib.h>
+#include <time.h>
 
 #include "expr-core-funcproc.h"
 #include "date.h"
 #include "statvar.h"
 #include "../core/nxlog.h"
 #include "../core/ctx.h"
+#include "alloc.h"
 
 #define NX_LOGMODULE NX_LOGMODULE_CORE
 
@@ -450,6 +452,8 @@ void nx_expr_func__to_string(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 			     int32_t num_arg,
 			     nx_value_t *args)
 {
+    char *str;
+
     ASSERT(retval != NULL);
     ASSERT(num_arg == 1);
 
@@ -460,8 +464,16 @@ void nx_expr_func__to_string(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 	return;
     }
 
-    retval->string = nx_string_create(nx_value_to_string(&(args[0])), -1);
-    retval->defined = TRUE;
+    str = nx_value_to_string(&(args[0]));
+    if ( str != NULL )
+    {
+	retval->string = nx_string_create_owned(str, -1);
+	retval->defined = TRUE;
+    }
+    else
+    {
+	retval->defined = FALSE;
+    }
 }
 
 
@@ -496,7 +508,6 @@ void nx_expr_func__to_integer(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 		  nx_value_type_to_string(args[0].type));
     }
     retval->defined = TRUE;
-
 }
 
 
@@ -551,9 +562,15 @@ void nx_expr_func__parsedate(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 	return;
     }
 
-    CHECKERR_MSG(nx_date_parse(&(retval->datetime), args[0].string->buf, NULL), 
-		 "couldn't parse date: %s", args[0].string->buf);
-    retval->defined = TRUE;
+    if ( nx_date_parse(&(retval->datetime), args[0].string->buf, NULL) == APR_SUCCESS )
+    {
+	retval->defined = TRUE;
+    }
+    else
+    {
+	log_debug("couldn't parse date: %s", args[0].string->buf);
+	retval->defined = FALSE;
+    }
 }
 
 
@@ -710,13 +727,112 @@ void nx_expr_func__hostname_fqdn(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 
 
 
+void nx_expr_func__host_ip(nx_expr_eval_ctx_t *eval_ctx UNUSED,
+			   nx_module_t *module UNUSED,
+			   nx_value_t *retval,
+			   int32_t num_arg,
+			   nx_value_t *args)
+{
+    nxlog_t *nxlog;
+    nx_exception_t e;
+    static apr_sockaddr_t *sa = NULL; // this holds a cached sa so that we don't perform a DNS lookup every time
+    volatile uint8_t ip4addr[4];
+    uint32_t ip4;
+    apr_sockaddr_t *_sa;
+    volatile int64_t nth = 1;
+    int64_t i = 1;
+
+    ASSERT(retval != NULL);
+
+    if ( num_arg == 1 )
+    {
+	if ( args[0].type != NX_VALUE_TYPE_INTEGER )
+	{
+	    throw_msg("got '%s' for first argument of function 'host_ip(integer)'",
+		      nx_value_type_to_string(args[0].type));
+	}
+	if ( args[0].defined == FALSE )
+	{
+	    retval->defined = FALSE;
+	    return;
+	}
+	nth = args[0].integer;
+	if ( nth <= 0 )
+	{
+	    throw_msg("Positive non-zero number required for argument nth in function 'host_ip(integer)', got %d", nth);
+	}
+    }
+    else
+    {
+	ASSERT(num_arg == 0);
+    }
+
+
+    retval->type = NX_VALUE_TYPE_IP4ADDR;
+    retval->defined = FALSE;
+
+    if ( sa == NULL )
+    { 
+	nx_lock();
+	try
+	{
+	    nxlog = nxlog_get();
+
+	    CHECKERR_MSG(apr_sockaddr_info_get(&_sa, nxlog->hostname.buf, APR_INET, 0, 0, nxlog->pool), 
+			 "cannot resolve hostname to ip address");
+	    sa = _sa;
+	}
+	catch(e)
+	{
+	    nx_unlock();
+	    rethrow(e);
+	}
+	nx_unlock();
+    }
+    else
+    {
+	_sa = sa;
+    }
+    while ( _sa != NULL )
+    {
+	if ( _sa->family == AF_INET )
+	{ // we are only interested in ipv4 addresses
+	    ip4 = ntohl(_sa->sa.sin.sin_addr.s_addr);
+	    ip4addr[3] = (uint8_t) (ip4 % 256);
+	    ip4addr[2] = (uint8_t) ((ip4 / 256) % 256);
+	    ip4addr[1] = (uint8_t) ((ip4 / (256 * 256)) % 256);
+	    ip4addr[0] = (uint8_t) ((ip4 / (256 * 256 * 256)) % 256);
+	    if ( ip4addr[0] != 127 )
+	    { // ignore loopback address
+		if (i == nth)
+		{
+		    retval->ip4addr[0] = ip4addr[0];
+		    retval->ip4addr[1] = ip4addr[1];
+		    retval->ip4addr[2] = ip4addr[2];
+		    retval->ip4addr[3] = ip4addr[3];
+		    retval->defined = TRUE;
+		    break; // return nth
+		}
+		i++;
+	    }
+	}
+	else
+	{
+	    log_info("non AF_INET");
+	}
+	_sa = _sa->next;
+    }
+}
+
+
+
 static void nx_expr_proc__log(nx_expr_eval_ctx_t *eval_ctx,
 			      nx_module_t *module UNUSED,
-			      nx_expr_arg_list_t *args,
+			      nx_expr_list_t *args,
 			      nx_loglevel_t loglevel)
 {
     int i, j;
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     char *buf;
     char **argvalues;
     char *tmpstr;
@@ -766,7 +882,7 @@ static void nx_expr_proc__log(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__log_debug(nx_expr_eval_ctx_t *eval_ctx,
 			     nx_module_t *module,
-			     nx_expr_arg_list_t *args)
+			     nx_expr_list_t *args)
 {
     nx_expr_proc__log(eval_ctx, module, args, NX_LOGLEVEL_DEBUG);
 }
@@ -775,7 +891,7 @@ void nx_expr_proc__log_debug(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__debug(nx_expr_eval_ctx_t *eval_ctx,
 			 nx_module_t *module,
-			 nx_expr_arg_list_t *args)
+			 nx_expr_list_t *args)
 {
     nx_expr_proc__log(eval_ctx, module, args, NX_LOGLEVEL_DEBUG);
 }
@@ -784,7 +900,7 @@ void nx_expr_proc__debug(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__log_info(nx_expr_eval_ctx_t *eval_ctx,
 			     nx_module_t *module,
-			     nx_expr_arg_list_t *args)
+			     nx_expr_list_t *args)
 {
     nx_expr_proc__log(eval_ctx, module, args, NX_LOGLEVEL_INFO);
 }
@@ -793,7 +909,7 @@ void nx_expr_proc__log_info(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__log_warning(nx_expr_eval_ctx_t *eval_ctx,
 			     nx_module_t *module,
-			     nx_expr_arg_list_t *args)
+			     nx_expr_list_t *args)
 {
     nx_expr_proc__log(eval_ctx, module, args, NX_LOGLEVEL_WARNING);
 }
@@ -802,7 +918,7 @@ void nx_expr_proc__log_warning(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__log_error(nx_expr_eval_ctx_t *eval_ctx,
 			     nx_module_t *module,
-			     nx_expr_arg_list_t *args)
+			     nx_expr_list_t *args)
 {
     nx_expr_proc__log(eval_ctx, module, args, NX_LOGLEVEL_ERROR);
 }
@@ -818,6 +934,8 @@ void nx_expr_func__replace(nx_expr_eval_ctx_t *eval_ctx UNUSED,
     int64_t count = 0;
     nx_string_t *result = NULL;
     const char *str, *ptr;
+    const char *dstptr;
+    int dstlen;
     int i;
 
     ASSERT(retval != NULL);
@@ -825,34 +943,42 @@ void nx_expr_func__replace(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 
     retval->type = NX_VALUE_TYPE_STRING;
 
+    if ( args[0].defined == FALSE )
+    {
+	return;
+    }
     if ( args[0].type != NX_VALUE_TYPE_STRING )
     {
 	throw_msg("'%s' type argument is invalid for argument 'subject', string required",
 		  nx_value_type_to_string(args[0].type));
     }
-    if ( args[0].defined == FALSE )
-    {
+
+    if ( args[1].defined != TRUE )
+    { // if 'src' is undef we don't modify 'subject'
+	retval->string = nx_string_clone(args[0].string);
+	retval->defined = TRUE;
 	return;
     }
-
     if ( args[1].type != NX_VALUE_TYPE_STRING )
     {
 	throw_msg("'%s' type argument is invalid for argument 'src', string required",
 		  nx_value_type_to_string(args[1].type));
     }
-    if ( args[1].defined != TRUE )
-    {
-	return;
-    }
 
-    if ( args[2].type != NX_VALUE_TYPE_STRING )
+    if ( args[2].defined != TRUE ) 
+    {
+	dstptr = "";
+	dstlen = 0;
+    }
+    else if ( args[2].type != NX_VALUE_TYPE_STRING )
     {
 	throw_msg("'%s' type argument is invalid for argument 'dst', string required",
 		  nx_value_type_to_string(args[2].type));
     }
-    if ( args[2].defined != TRUE )
-    {
-	return;
+    else
+    { // NX_VALUE_TYPE_STRING 
+	dstptr = args[2].string->buf;
+	dstlen = (int) args[2].string->len;
     }
 
     if ( num_arg == 4 )
@@ -897,7 +1023,7 @@ void nx_expr_func__replace(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 	    result = nx_string_new_size(args[0].string->len);
 	}
 	nx_string_append(result, str, (int) (ptr - str));
-	nx_string_append(result, args[2].string->buf, (int) args[2].string->len);
+	nx_string_append(result, dstptr, dstlen);
 	str = ptr + args[1].string->len;
     }
     if ( str - args[0].string->buf < args[0].string->len )
@@ -920,7 +1046,7 @@ void nx_expr_func__replace(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 
 void nx_expr_proc__drop(nx_expr_eval_ctx_t *eval_ctx,
 			nx_module_t *module UNUSED,
-			nx_expr_arg_list_t *args UNUSED)
+			nx_expr_list_t *args UNUSED)
 {
     if ( eval_ctx->logdata == NULL )
     {
@@ -928,21 +1054,22 @@ void nx_expr_proc__drop(nx_expr_eval_ctx_t *eval_ctx,
     }
     // must not free it here because it is still linked in the queue
     eval_ctx->logdata = NULL;
+    eval_ctx->dropped = TRUE; // don't continue executing further statements
 }
 
 
 
 void nx_expr_proc__delete(nx_expr_eval_ctx_t *eval_ctx,
 			  nx_module_t *module UNUSED,
-			  nx_expr_arg_list_t *args)
+			  nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
 
     ASSERT(args != NULL);
 
     if ( eval_ctx->logdata == NULL )
     {
-	throw_msg("no logdata to delete() from, possbily dropped");
+	throw_msg("no logdata to delete() from, possibly dropped");
     }
 
     arg = NX_DLIST_FIRST(args);
@@ -963,9 +1090,9 @@ void nx_expr_proc__delete(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__create_var(nx_expr_eval_ctx_t *eval_ctx,
 			      nx_module_t *module,
-			      nx_expr_arg_list_t *args)
+			      nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t argval1, argval2;
     apr_time_t expiry = 0;
 
@@ -1034,9 +1161,9 @@ void nx_expr_proc__create_var(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__delete_var(nx_expr_eval_ctx_t *eval_ctx,
 			      nx_module_t *module UNUSED,
-			      nx_expr_arg_list_t *args)
+			      nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t argval;
     nx_exception_t e;
 
@@ -1069,10 +1196,10 @@ void nx_expr_proc__delete_var(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__set_var(nx_expr_eval_ctx_t *eval_ctx,
 			   nx_module_t *module,
-			   nx_expr_arg_list_t *args)
+			   nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg1;
-    nx_expr_arg_t *arg2;
+    nx_expr_list_elem_t *arg1;
+    nx_expr_list_elem_t *arg2;
     nx_value_t argval;
     nx_module_var_t *var;
     nx_value_t tmpval;
@@ -1178,9 +1305,9 @@ void nx_expr_func__get_var(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__create_stat(nx_expr_eval_ctx_t *eval_ctx,
 			       nx_module_t *module,
-			       nx_expr_arg_list_t *args)
+			       nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t argval1, argval2, argval3;
     apr_time_t expiry = 0;
     int64_t interval = 0;
@@ -1333,11 +1460,11 @@ void nx_expr_proc__create_stat(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__add_stat(nx_expr_eval_ctx_t *eval_ctx,
 			    nx_module_t *module,
-			    nx_expr_arg_list_t *args)
+			    nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg1;
-    nx_expr_arg_t *arg2;
-    nx_expr_arg_t *arg3;
+    nx_expr_list_elem_t *arg1;
+    nx_expr_list_elem_t *arg2;
+    nx_expr_list_elem_t *arg3;
     nx_value_t argval1, argval2;
     nx_module_stat_t *stat;
     apr_time_t timeval = 0;
@@ -1481,9 +1608,9 @@ void nx_expr_func__get_stat(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__sleep(nx_expr_eval_ctx_t *eval_ctx,
 			      nx_module_t *module UNUSED,
-			      nx_expr_arg_list_t *args)
+			      nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t argval;
 
     ASSERT(args != NULL);
@@ -1505,9 +1632,9 @@ void nx_expr_proc__sleep(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__rename_field(nx_expr_eval_ctx_t *eval_ctx,
 				nx_module_t *module UNUSED,
-				nx_expr_arg_list_t *args)
+				nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t argval1, argval2;
     nx_exception_t e;
 
@@ -1543,7 +1670,7 @@ void nx_expr_proc__rename_field(nx_expr_eval_ctx_t *eval_ctx,
     }
     if ( argval2.string->len == 0 )
     {
-	nx_value_kill(&argval2);
+	nx_value_kill(&argval1);
 	nx_value_kill(&argval2);
 	throw_msg("zero sized string is invalid for second arg of rename_field()");
     }
@@ -1554,12 +1681,12 @@ void nx_expr_proc__rename_field(nx_expr_eval_ctx_t *eval_ctx,
     }
     catch(e)
     {
-	nx_value_kill(&argval2);
+	nx_value_kill(&argval1);
 	nx_value_kill(&argval2);
 	rethrow(e);
     }
 
-    nx_value_kill(&argval2);
+    nx_value_kill(&argval1);
     nx_value_kill(&argval2);
 }
 
@@ -1567,9 +1694,9 @@ void nx_expr_proc__rename_field(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__reroute(nx_expr_eval_ctx_t *eval_ctx,
 			   nx_module_t *module,
-			   nx_expr_arg_list_t *args)
+			   nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t routeval;
     nx_logdata_t *logdata;
     nx_ctx_t *ctx;
@@ -1608,9 +1735,14 @@ void nx_expr_proc__reroute(nx_expr_eval_ctx_t *eval_ctx,
 
     if ( route == NULL )
     {
-	throw_msg("invalid route name specified for reroute(): '%s'", routeval.string->buf);
+	char tmpstr[200];
+
+	apr_cpystrn(tmpstr, routeval.string->buf, sizeof(tmpstr));
+	nx_value_kill(&routeval);
+	throw_msg("invalid route name specified for reroute(): '%s'", tmpstr);
     }
 
+    nx_value_kill(&routeval);
     nx_module_add_logdata_to_route(module, route, logdata);
 
     // must not free it here because it is still linked in the queue
@@ -1621,9 +1753,9 @@ void nx_expr_proc__reroute(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__add_to_route(nx_expr_eval_ctx_t *eval_ctx,
 				nx_module_t *module,
-				nx_expr_arg_list_t *args)
+				nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t routeval;
     nx_logdata_t *logdata;
     nx_ctx_t *ctx;
@@ -1661,8 +1793,14 @@ void nx_expr_proc__add_to_route(nx_expr_eval_ctx_t *eval_ctx,
 
     if ( route == NULL )
     {
-	throw_msg("invalid route name specified for add_to_route(): '%s'", routeval.string->buf);
+	char tmpstr[200];
+
+	apr_cpystrn(tmpstr, routeval.string->buf, sizeof(tmpstr));
+	nx_value_kill(&routeval);
+
+	throw_msg("invalid route name specified for add_to_route(): '%s'", tmpstr);
     }
+    nx_value_kill(&routeval);
 
     nx_module_lock(module);
     (module->evt_fwd)++;
