@@ -38,29 +38,34 @@ void nx_expr_parser_error(nx_expr_parser_t *parser UNUSED,
 
 
 
-//TODO: use pool alloc
-void nx_expr_parser_append_string(char **dst,
+void nx_expr_parser_append_string(nx_expr_parser_t *parser,
+				  char **dst,
 				  const char *src)
 {
     size_t len1, len2;
 
+    ASSERT(parser != NULL);
     ASSERT(src != NULL);
 
-    log_debug("appending [%s] to [%s]", src, *dst);
     if ( *dst == NULL )
     {
+	log_debug("adding string [%s]", src);
 	len1 = strlen(src);
-	*dst = malloc(len1 + 1);
+	*dst = apr_palloc(parser->pool, len1 + 1);
 	apr_cpystrn(*dst, src, len1 + 1);
     }
     else
     {
+	char *tmp;
+	log_debug("appending [%s] to [%s]", src, *dst);
 	len1 = strlen(*dst);
 	len2 = strlen(src);
-	*dst = realloc(*dst, len1 + len2 + 1);
-	apr_cpystrn(*dst + len1, src, len2 + 1);
+
+	tmp = apr_palloc(parser->pool, len1 + len2 + 1);
+	apr_cpystrn(tmp, *dst, len1 + 1);
+	apr_cpystrn(tmp + len1, src, len2 + 1);
+	*dst = tmp;
     }
-	   
 }
 
 
