@@ -26,7 +26,7 @@ typedef union nx_expr_parser_yystype
     char	*string;
     char	*regexp[3]; // regexp, replacement, options
     boolean	bool;
-    nx_expr_arg_list_t *args;
+    nx_expr_list_t *exprs;
     nx_expr_statement_list_t *statement_list;
     nx_expr_statement_t *statement;
 } nx_expr_parser_yystype;
@@ -54,7 +54,9 @@ void nx_expr_parser_error_fmt(nx_expr_parser_t	*parser,
 void nx_expr_parser_error(nx_expr_parser_t *parser,
 			  void *scanner UNUSED,
 			  const char *msg) NORETURN;
-void nx_expr_parser_append_string(char **dst, const char *src);
+void nx_expr_parser_append_string(nx_expr_parser_t *parser,
+				  char **dst,
+				  const char *src);
 void nx_expr_parser_unescape_string(char *str);
 const char *nx_expr_parser_new_string(nx_expr_parser_t *parser, const char *src);
 
