@@ -18,6 +18,11 @@
 
 #define NX_LOGMODULE NX_LOGMODULE_CORE
 
+static void _regmatch(nx_expr_eval_ctx_t *eval_ctx,
+		      nx_value_t *retval,
+		      const nx_expr_t *left,
+		      const nx_expr_t *right);
+
 
 static void nx_expr_decl_init(nx_expr_decl_t *decl, 
 			      nx_expr_parser_t *parser,
@@ -31,10 +36,11 @@ static void nx_expr_decl_init(nx_expr_decl_t *decl,
     {
 	decl->file = "INPUT";
     }
+
     decl->line = parser->linenum;
     decl->pos = parser->linepos;
     decl->name = exprname;
-    //log_debug("%s declared at line %d, character %d", decl->name, decl->line, decl->pos);
+    log_debug("%s declared at line %d, character %d in %s", decl->name, decl->line, decl->pos, decl->file);
 }
 
 
@@ -136,6 +142,10 @@ void nx_expr_statement_execute(nx_expr_eval_ctx_t *eval_ctx,
     ASSERT(eval_ctx != NULL);
     ASSERT(stmnt != NULL);
 
+    if ( eval_ctx->dropped == TRUE )
+    {
+	return;
+    }
 
     switch ( stmnt->type )
     {
@@ -185,6 +195,38 @@ void nx_expr_statement_execute(nx_expr_eval_ctx_t *eval_ctx,
 		nx_expr_statement_list_execute(eval_ctx, stmnt->block.statements);
 	    }
 	    break;
+	case NX_EXPR_STATEMENT_TYPE_REGEXPREPLACE:
+	    try
+	    {
+		nx_value_t retval;
+		_regmatch(eval_ctx, &retval, stmnt->regexpreplace.lval,
+			  stmnt->regexpreplace.rval);
+		nx_value_kill(&retval);
+	    }
+	    catch(e)
+	    {
+		rethrow_msg(e, "%s failed at line %d, character %d in %s. "
+			    "statement execution has been aborted",
+			    stmnt->decl.name, stmnt->decl.line, stmnt->decl.pos,
+			    stmnt->decl.file);
+	    }
+	    break;
+	case NX_EXPR_STATEMENT_TYPE_REGEXP:
+	    try
+	    {
+		nx_value_t retval;
+		_regmatch(eval_ctx, &retval, stmnt->regexp.lval,
+			  stmnt->regexp.rval);
+		nx_value_kill(&retval);
+	    }
+	    catch(e)
+	    {
+		rethrow_msg(e, "%s failed at line %d, character %d in %s. "
+			    "statement execution has been aborted",
+			    stmnt->decl.name, stmnt->decl.line, stmnt->decl.pos,
+			    stmnt->decl.file);
+	    }
+	    break;
 	default:
 	    nx_panic("invalid statement type %d", stmnt->type);
     }
@@ -198,7 +240,7 @@ static void nx_expr_eval_func(nx_expr_eval_ctx_t *eval_ctx,
 {
     int volatile argc, i;
     nx_value_t *values;
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_exception_t e;
     
     if ( expr->function.args == NULL )
@@ -261,6 +303,7 @@ void nx_expr_eval_ctx_init(nx_expr_eval_ctx_t *eval_ctx,
     eval_ctx->input = input;
     eval_ctx->num_captured = 0;
     eval_ctx->captured = NULL;
+    eval_ctx->dropped = FALSE;
 }
 
 
@@ -283,6 +326,85 @@ void nx_expr_eval_ctx_destroy(nx_expr_eval_ctx_t *eval_ctx)
 
 
 
+static void _inop(nx_expr_eval_ctx_t *eval_ctx,
+		  nx_value_t *retval,
+		  const nx_expr_t *expr,
+		  const nx_expr_list_t *exprs)
+{
+    nx_value_t exprval;
+    nx_exception_t e;
+    nx_expr_list_elem_t *elem;
+    int volatile elemcount = 0;
+    nx_value_t *values = NULL;
+    int i = 0;
+
+    ASSERT(expr != NULL);
+    ASSERT(exprs != NULL);
+
+    retval->type = NX_VALUE_TYPE_BOOLEAN;
+    retval->defined = TRUE;
+    retval->boolean = FALSE;
+
+    for ( elem = NX_DLIST_FIRST(exprs), elemcount = 0;
+	  elem != NULL;
+	  elem = NX_DLIST_NEXT(elem, link), elemcount++ );
+
+    values = malloc((size_t) elemcount * sizeof(nx_value_t));
+    memset(values, 0, (size_t) elemcount * sizeof(nx_value_t));
+	
+    nx_expr_evaluate(eval_ctx, &exprval, expr);
+
+    if ( exprval.defined == FALSE )
+    {
+	// evaluate to UNDEF if expr is undef
+	retval->defined = FALSE;
+	return;
+    }
+
+    try
+    {
+	for ( elem = NX_DLIST_FIRST(exprs), i = 0;
+	      elem != NULL;
+	      elem = NX_DLIST_NEXT(elem, link), i++ )
+	{
+	    nx_expr_evaluate(eval_ctx, &(values[i]), elem->expr);
+
+	    if ( nx_value_eq(&exprval, &(values[i])) == TRUE )
+	    {
+		retval->boolean = TRUE;
+		break;
+	    }
+	}
+    }
+    catch(e)
+    {
+	for ( i = 0; i < elemcount; i++ )
+	{
+	    nx_value_kill(&(values[i]));
+	}
+	nx_value_kill(&exprval);
+	if ( values != NULL )
+	{
+	    free(values);
+	}
+	rethrow(e);
+    }
+
+    for ( i = 0; i < elemcount; i++ )
+    {
+	nx_value_kill(&(values[i]));
+    }
+    nx_value_kill(&exprval);
+    if ( values != NULL )
+    {
+	free(values);
+    }
+
+    return;
+}
+
+
+
 void nx_expr_evaluate(nx_expr_eval_ctx_t *eval_ctx,
 		      nx_value_t *retval,
 		      const nx_expr_t *expr)
@@ -359,6 +481,20 @@ void nx_expr_evaluate(nx_expr_eval_ctx_t *eval_ctx,
 			    expr->decl.file);
 	    }
 	    break;
+	case NX_EXPR_TYPE_INOP:
+	    log_debug("evaluating in");
+	    try
+	    {
+		_inop(eval_ctx, retval, expr->inop.expr, expr->inop.exprs);
+	    }
+	    catch(e)
+	    {
+		rethrow_msg(e, "%s failed at line %d, character %d in %s. "
+			    "expression evaluation has been aborted",
+			    expr->decl.name, expr->decl.line, expr->decl.pos,
+			    expr->decl.file);
+	    }
+	    break;
 	case NX_EXPR_TYPE_FUNCTION:
 	    log_debug("evaluating function");
 	    try
@@ -397,6 +533,7 @@ static void _regmatch(nx_expr_eval_ctx_t *eval_ctx,
     int ovector[NX_EXPR_MAX_CAPTURED_FIELDS * 3];
     nx_value_t fieldval;
     int global_match_recursion_cnt = 0;
+    int options = 0;
 
     ASSERT(retval != NULL);
     ASSERT(left != NULL);
@@ -489,7 +626,7 @@ static void _regmatch(nx_expr_eval_ctx_t *eval_ctx,
     retval->defined = TRUE;
 
   match_global:
-    result = pcre_exec(regexp, NULL, subject, (int) strlen(subject), 0, 0,
+    result = pcre_exec(regexp, NULL, subject, (int) strlen(subject), 0, options,
 		       ovector, NX_EXPR_MAX_CAPTURED_FIELDS * 3);
     
     if ( result >= 0 )
@@ -538,8 +675,21 @@ static void _regmatch(nx_expr_eval_ctx_t *eval_ctx,
     {
 	ASSERT(result < NX_EXPR_MAX_CAPTURED_FIELDS);
 
+	// clear captured strings from pervious operation
+	for ( i = 0; i < eval_ctx->num_captured; i++ )
+	{
+	    nx_string_free(eval_ctx->captured[i]);
+	}
+	eval_ctx->num_captured = 0;
+	if ( eval_ctx->captured != NULL )
+	{
+	    free(eval_ctx->captured);
+	    eval_ctx->captured = NULL;
+	}
+
 	eval_ctx->num_captured = result;
 	eval_ctx->captured = malloc((size_t) eval_ctx->num_captured * sizeof(nx_string_t *));
+
 	for ( i = 0; i < result; i++ )
 	{ 
 	    eval_ctx->captured[i] = nx_string_create(subject + ovector[i * 2],
@@ -667,6 +817,11 @@ static void _equal(nx_expr_eval_ctx_t *eval_ctx,
 	case NX_VALUE_TYPE_DATETIME:
 	    retval->boolean = lval.datetime == rval.datetime;
 	    break;
+	case NX_VALUE_TYPE_IP4ADDR:
+	    retval->boolean =
+		(lval.ip4addr[0] == rval.ip4addr[0]) && (lval.ip4addr[1] == rval.ip4addr[1]) && 
+		(lval.ip4addr[2] == rval.ip4addr[2]) && (lval.ip4addr[3] == rval.ip4addr[3]);
+    	    break;
 	default:
 	{
 	    nx_value_type_t rtype, ltype;
@@ -756,6 +911,11 @@ static void _notequal(nx_expr_eval_ctx_t *eval_ctx,
 	case NX_VALUE_TYPE_DATETIME:
 	    retval->boolean = lval.datetime != rval.datetime;
 	    break;
+	case NX_VALUE_TYPE_IP4ADDR:
+	    retval->boolean =
+		(lval.ip4addr[0] != rval.ip4addr[0]) || (lval.ip4addr[1] != rval.ip4addr[1]) ||
+		(lval.ip4addr[2] != rval.ip4addr[2]) || (lval.ip4addr[3] != rval.ip4addr[3]);
+    	    break;
 	default:
 	{
 	    nx_value_type_t rtype, ltype;
@@ -1063,6 +1223,12 @@ static void _and(nx_expr_eval_ctx_t *eval_ctx,
 	retval->defined = FALSE;
 	return;
     }
+    else if ( (lval.type == NX_VALUE_TYPE_BOOLEAN) && (lval.boolean == FALSE) )
+    { // short-circuit if left operand is false
+	retval->defined = TRUE;
+	retval->boolean = FALSE;
+	return;
+    }
 
     nx_expr_evaluate(eval_ctx, &rval, right);
     if ( rval.defined == FALSE )
@@ -2019,6 +2185,24 @@ nx_expr_t *nx_expr_new_unop(nx_expr_parser_t *parser,
 
 
 
+nx_expr_t *nx_expr_new_inop(nx_expr_parser_t *parser,
+			    const nx_expr_t *expr,
+			    const nx_expr_list_t *exprs)
+{
+    nx_expr_t *retval;
+    
+    retval = apr_pcalloc(parser->pool, sizeof(nx_expr_t));
+    retval->type = NX_EXPR_TYPE_INOP;
+    retval->rettype = NX_VALUE_TYPE_BOOLEAN;
+    retval->inop.expr = expr;
+    retval->inop.exprs = exprs;
+    nx_expr_decl_init(&(retval->decl), parser, "in");
+
+    return ( retval );
+}
+
+
+
 nx_expr_t *nx_expr_new_string(nx_expr_parser_t *parser, const char *str)
 {
     nx_expr_t *expr;
@@ -2072,9 +2256,11 @@ nx_expr_t *nx_expr_new_regexp(nx_expr_parser_t *parser,
     pcre *regexp;
     pcre *tmppcre;
     int rc;
-    int size;
+    size_t size;
     int capturecount;
     uint8_t mod = 0;
+    size_t i;
+    int options = 0;
 
     ASSERT(str != NULL);
 
@@ -2089,22 +2275,44 @@ nx_expr_t *nx_expr_new_regexp(nx_expr_parser_t *parser,
 
     if ( modifiers != NULL )
     {
-	if ( replacement == NULL )
+	for ( i = 0; i < strlen(modifiers); i++ )
 	{
-	    nx_expr_parser_error_fmt(parser, NULL,
-				     "regular expression modifier is invalid: %s",
-				     modifiers);
+	    switch ( modifiers[i] )
+	    {
+		case 'g':
+		    mod |= NX_EXPR_REGEXP_MODIFIER_MATCH_GLOBAL;
+		    break;
+		case 's':
+		    mod |= NX_EXPR_REGEXP_MODIFIER_MATCH_DOTALL;
+		    break;
+		case 'm':
+		    mod |= NX_EXPR_REGEXP_MODIFIER_MATCH_MULTILINE;
+		    break;
+		case 'i':
+		    mod |= NX_EXPR_REGEXP_MODIFIER_MATCH_CASELESS;
+		    break;
+		default:
+		    nx_expr_parser_error_fmt(parser, NULL,
+					     "regular expression modifier is invalid: %c",
+					     modifiers[i]);
+	    }
 	}
-	if ( strcmp(modifiers, "g") != 0 )
-	{
-	    nx_expr_parser_error_fmt(parser, NULL,
-				     "regular expression modifier is invalid: %s",
-				     modifiers);
-	}
-	mod = NX_EXPR_REGEXP_MODIFIER_MATCH_GLOBAL;
     }
 
-    regexp = pcre_compile(str, 0, &error,  &erroroffs, NULL);
+    if ( mod & NX_EXPR_REGEXP_MODIFIER_MATCH_DOTALL )
+    {
+	options |= PCRE_DOTALL;
+    }
+    if ( mod & NX_EXPR_REGEXP_MODIFIER_MATCH_MULTILINE )
+    {
+	options |= PCRE_MULTILINE;
+    }
+    if ( mod & NX_EXPR_REGEXP_MODIFIER_MATCH_CASELESS )
+    {
+	options |= PCRE_CASELESS;
+    }
+
+    regexp = pcre_compile(str, options, &error,  &erroroffs, NULL);
 
     if ( regexp == NULL )
     {
@@ -2147,7 +2355,7 @@ nx_expr_t *nx_expr_new_regexp(nx_expr_parser_t *parser,
 	expr->value.type = NX_VALUE_TYPE_REGEXP;
 	expr->value.defined = TRUE;
 	expr->value.regexp.pcre = tmppcre;
-	expr->value.regexp.pcre_size = (size_t) size;
+	expr->value.regexp.pcre_size = size;
 	expr->value.regexp.str = apr_pstrdup(parser->pool, str);
 	expr->value.regexp.replacement = apr_pstrdup(parser->pool, replacement);
 	expr->value.regexp.modifiers = mod;
@@ -2160,10 +2368,10 @@ nx_expr_t *nx_expr_new_regexp(nx_expr_parser_t *parser,
 	expr->value.type = NX_VALUE_TYPE_REGEXP;
 	expr->value.defined = TRUE;
 	expr->value.regexp.pcre = tmppcre;
-	expr->value.regexp.pcre_size = (size_t) size;
+	expr->value.regexp.pcre_size = size;
 	expr->value.regexp.str = apr_pstrdup(parser->pool, str);
 	expr->value.regexp.replacement = NULL;
-	expr->value.regexp.modifiers = 0;
+	expr->value.regexp.modifiers = mod;
 	nx_expr_decl_init(&(expr->decl), parser, "regexp");
     }
 
@@ -2274,12 +2482,29 @@ nx_expr_t *nx_expr_new_datetime(nx_expr_parser_t *parser, const char *str)
 
 
 
-static boolean nx_expr_args_match(const nx_expr_arg_list_t *args,
-				  int num_arg,
-				  nx_value_type_t *arg_types)
+nx_expr_t *nx_expr_new_ip4addr(nx_expr_parser_t *parser, const char *str)
+{
+    nx_expr_t *expr;
+    
+    expr = apr_pcalloc(parser->pool, sizeof(nx_expr_t));
+    expr->type = NX_EXPR_TYPE_VALUE;
+    nx_value_parse_ip4addr(&(expr->value), str);
+    expr->value.type = NX_VALUE_TYPE_IP4ADDR;
+    expr->value.defined = TRUE;
+    expr->rettype = NX_VALUE_TYPE_IP4ADDR;
+    nx_expr_decl_init(&(expr->decl), parser, "ip4addr literal");
+
+    return ( expr );
+}
+
+
+
+static boolean nx_expr_list_elem_match(const nx_expr_list_t *args,
+				       int num_arg,
+				       nx_value_type_t *arg_types)
 {
     boolean retval = TRUE;
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     int i;
 
     if ( (args == NULL) || (NX_DLIST_FIRST(args) == NULL) )
@@ -2324,7 +2549,7 @@ static boolean nx_expr_args_match(const nx_expr_arg_list_t *args,
 
 nx_expr_t *nx_expr_new_function(nx_expr_parser_t *parser,
 				const char *fname,
-				nx_expr_arg_list_t *args)
+				nx_expr_list_t *args)
 {
     nx_expr_t *expr;
     nx_ctx_t *ctx;
@@ -2381,7 +2606,7 @@ nx_expr_t *nx_expr_new_function(nx_expr_parser_t *parser,
 	    if ( (func->type == NX_EXPR_FUNCPROC_TYPE_PRIVATE) &&
 		 (parser->module == func->module) &&
 		 (strcmp(func->name, fname) == 0) &&
-		 (nx_expr_args_match(args, func->num_arg, func->arg_types) == TRUE) )
+		 (nx_expr_list_elem_match(args, func->num_arg, func->arg_types) == TRUE) )
 	    {
 		ASSERT(func->module != NULL);
 		found = TRUE;
@@ -2402,7 +2627,7 @@ nx_expr_t *nx_expr_new_function(nx_expr_parser_t *parser,
 	{
 	    //log_info("%s <=> %s", func->name, fname);
 	    if ( (strcmp(func->name, fname) == 0) &&
-		 (nx_expr_args_match(args, func->num_arg, func->arg_types) == TRUE) )
+		 (nx_expr_list_elem_match(args, func->num_arg, func->arg_types) == TRUE) )
 	    {
 		switch ( func->type )
 		{
@@ -2601,17 +2826,17 @@ void nx_expr_proc_register(apr_pool_t		*pool,
 
 
 
-nx_expr_arg_list_t *nx_expr_arg_list_new(nx_expr_parser_t *parser,
-					 nx_expr_t *expr)
+nx_expr_list_t *nx_expr_list_new(nx_expr_parser_t *parser,
+				 nx_expr_t *expr)
 {
-    nx_expr_arg_list_t *retval;
-    nx_expr_arg_t *arg;
+    nx_expr_list_t *retval;
+    nx_expr_list_elem_t *arg;
 
     ASSERT(expr != NULL);
 
-    retval = apr_pcalloc(parser->pool, sizeof(nx_expr_arg_list_t));
-    NX_DLIST_INIT(retval, nx_expr_arg_t, link);
-    arg = apr_pcalloc(parser->pool, sizeof(nx_expr_arg_t));
+    retval = apr_pcalloc(parser->pool, sizeof(nx_expr_list_t));
+    NX_DLIST_INIT(retval, nx_expr_list_elem_t, link);
+    arg = apr_pcalloc(parser->pool, sizeof(nx_expr_list_elem_t));
     arg->expr = expr;
    
     NX_DLIST_INSERT_TAIL(retval, arg, link);
@@ -2621,16 +2846,16 @@ nx_expr_arg_list_t *nx_expr_arg_list_new(nx_expr_parser_t *parser,
 
 
 
-nx_expr_arg_list_t *nx_expr_arg_list_add(nx_expr_parser_t *parser,
-					 nx_expr_arg_list_t *args,
-					 nx_expr_t *expr)
+nx_expr_list_t *nx_expr_list_add(nx_expr_parser_t *parser,
+				 nx_expr_list_t *args,
+				 nx_expr_t *expr)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
 
     ASSERT(args != NULL);
     ASSERT(expr != NULL);
 
-    arg = apr_pcalloc(parser->pool, sizeof(nx_expr_arg_t));
+    arg = apr_pcalloc(parser->pool, sizeof(nx_expr_list_elem_t));
     arg->expr = expr;
    
     NX_DLIST_INSERT_TAIL(args, arg, link);
@@ -2700,7 +2925,7 @@ nx_expr_statement_t *nx_expr_statement_new_block(nx_expr_parser_t *parser,
 
 nx_expr_statement_t *nx_expr_statement_new_procedure(nx_expr_parser_t *parser,
 						     const char *name,
-						     nx_expr_arg_list_t *args)
+						     nx_expr_list_t *args)
 {
     nx_ctx_t *ctx;
     const char *tmp = NULL;
@@ -2760,7 +2985,7 @@ nx_expr_statement_t *nx_expr_statement_new_procedure(nx_expr_parser_t *parser,
 	    if ( (proc->type == NX_EXPR_FUNCPROC_TYPE_PRIVATE) &&
 		 (parser->module == proc->module) &&
 		 (strcmp(proc->name, name) == 0) &&
-		 (nx_expr_args_match(args, proc->num_arg, proc->arg_types) == TRUE) )
+		 (nx_expr_list_elem_match(args, proc->num_arg, proc->arg_types) == TRUE) )
 	    {
 		ASSERT(proc->module != NULL);
 		found = TRUE;
@@ -2783,7 +3008,7 @@ nx_expr_statement_t *nx_expr_statement_new_procedure(nx_expr_parser_t *parser,
 		      proc->module == NULL ? "NULL" : proc->module->name,
 		      module == NULL ? "NULL" : module->name);
 	    if ( (strcmp(proc->name, name) == 0) &&
-		 (nx_expr_args_match(args, proc->num_arg, proc->arg_types) == TRUE) )
+		 (nx_expr_list_elem_match(args, proc->num_arg, proc->arg_types) == TRUE) )
 	    {
 		switch ( proc->type )
 		{
@@ -2850,17 +3075,57 @@ nx_expr_statement_t *nx_expr_statement_new_procedure(nx_expr_parser_t *parser,
 
 
 
+nx_expr_statement_t *nx_expr_statement_new_regexpreplace(nx_expr_parser_t *parser,
+							 const nx_expr_t *lval,
+							 const nx_expr_t *rval)
+{
+    nx_expr_statement_t *retval;
+
+    ASSERT(lval != NULL);
+    ASSERT(rval != NULL);
+
+    retval = apr_pcalloc(parser->pool, sizeof(nx_expr_statement_t));
+    retval->type = NX_EXPR_STATEMENT_TYPE_REGEXPREPLACE;
+    retval->regexpreplace.lval = lval;
+    retval->regexpreplace.rval = rval;
+    nx_expr_decl_init(&(retval->decl), parser, "regexpreplace");
+
+    return ( retval );
+}
+
+
+
+nx_expr_statement_t *nx_expr_statement_new_regexp(nx_expr_parser_t *parser,
+						  const nx_expr_t *lval,
+						  const nx_expr_t *rval)
+{
+    nx_expr_statement_t *retval;
+
+    ASSERT(lval != NULL);
+    ASSERT(rval != NULL);
+
+    retval = apr_pcalloc(parser->pool, sizeof(nx_expr_statement_t));
+    retval->type = NX_EXPR_STATEMENT_TYPE_REGEXP;
+    retval->regexpreplace.lval = lval;
+    retval->regexpreplace.rval = rval;
+    nx_expr_decl_init(&(retval->decl), parser, "regexp");
+
+    return ( retval );
+}
+
+
+
 nx_expr_statement_list_t *nx_expr_statement_list_new(nx_expr_parser_t *parser,
 						     nx_expr_statement_t *stmnt)
 {
     nx_expr_statement_list_t *retval;
 
-    ASSERT(stmnt != NULL);
-
     retval = apr_pcalloc(parser->pool, sizeof(nx_expr_statement_list_t));
     NX_DLIST_INIT(retval, nx_expr_statement_t, link);
-    NX_DLIST_INSERT_TAIL(retval, stmnt, link);
-
+    if ( stmnt != NULL )
+    {
+	NX_DLIST_INSERT_TAIL(retval, stmnt, link);
+    }
     return ( retval );
 }
 
@@ -2870,10 +3135,12 @@ nx_expr_statement_list_t *nx_expr_statement_list_add(nx_expr_parser_t *parser UN
 						     nx_expr_statement_list_t *list,
 						     nx_expr_statement_t *stmnt)
 {
-    ASSERT(stmnt != NULL);
     ASSERT(list != NULL);
 
-    NX_DLIST_INSERT_TAIL(list, stmnt, link);
+    if ( stmnt != NULL )
+    {
+	NX_DLIST_INSERT_TAIL(list, stmnt, link);
+    }
 
     return ( list );
 }
