@@ -20,6 +20,8 @@
 
 #define NX_LOGMODULE NX_LOGMODULE_MODULE
 
+#define NX_POLLSET_REQEVENTS_KEY "reqevents"
+
 static nxlog_t *_nxlog = NULL;
 
 
@@ -433,7 +435,11 @@ boolean nx_module_can_send(nx_module_t *module, double multiplier)
 
 
 
-void nx_module_add_logdata_to_route(nx_module_t *module,
+/*
+ * This is only called from reroute() and add_to_route() procedures
+ * Flow Control is not supported here because it cannot be.
+ */
+void nx_module_add_logdata_to_route(nx_module_t *module UNUSED,
 				    nx_route_t *route,
 				    nx_logdata_t *logdata)
 {
@@ -447,30 +453,27 @@ void nx_module_add_logdata_to_route(nx_module_t *module,
 	
 	if ( curr->type == NX_MODULE_TYPE_PROCESSOR )
 	{
-	    tmp = nx_logdata_clone(logdata);
-	    if ( (nx_logqueue_push(curr->queue, tmp) >= curr->queue->limit) &&
-		 (nx_module_get_status(module) == NX_MODULE_STATUS_RUNNING) )
-	    {
-		nx_module_pause(module);
+	    if ( nx_logqueue_size(curr->queue) >= curr->queue->limit )
+	    { // cannot forward so we drop it since there is no flow-control
 	    }
 	    else
 	    {
+		tmp = nx_logdata_clone(logdata);
+		nx_logqueue_push(curr->queue, tmp);
 		nx_module_data_available(curr);
 	    }
-
 	    break; //skip remaining modules
 	}
 	
 	if ( curr->type == NX_MODULE_TYPE_OUTPUT )
 	{
-	    tmp = nx_logdata_clone(logdata);
-	    if ( (nx_logqueue_push(curr->queue, tmp) >= curr->queue->limit) &&
-		 (nx_module_get_status(module) == NX_MODULE_STATUS_RUNNING) )
-	    {
-		nx_module_pause(module);
+	    if ( nx_logqueue_size(curr->queue) >= curr->queue->limit )
+	    { // cannot forward so we drop it since there is no flow-control
 	    }
-	    if ( nx_module_can_send(curr, 1.0) == TRUE )
+	    else
 	    {
+		tmp = nx_logdata_clone(logdata);
+		nx_logqueue_push(curr->queue, tmp);
 		nx_module_data_available(curr);
 	    }
 	}
@@ -779,12 +782,15 @@ boolean nx_module_common_keyword(const char *keyword)
 {
     ASSERT(keyword != NULL);
 
+    // TODO: some directives are only for specific module types,
+    // e.g. FlowControl is not valid for output modules.
+
     if ( strcasecmp(keyword, "Module") == 0 )
     {
 	return ( TRUE );
     }
     if ( strcasecmp(keyword, "Processors") == 0 )
-    {
+    { //TODO: deprecated, remove it
 	return ( TRUE );
     }
     if ( strcasecmp(keyword, "Exec") == 0 )
@@ -1291,15 +1297,68 @@ void nx_module_shutdown(nx_module_t *module)
     nx_module_pollset_wakeup(module);
 }
 
+/* Get memory size on linux - for debugging
+ * Does not work well with APR_POOL_DEBUG
+ */
+/*
+#include <unistd.h>
+static int get_mem_size()
+{
+    int retval = -1;
+    apr_status_t rv;
+    apr_file_t *statfile = NULL;
+    apr_pool_t *pool;
+    char statbuf[1024];
+    apr_size_t nbytes;
+    int pid;
+    char comm[PATH_MAX + 2];
+    char state;
+    int ppid;
+    int pgrp;
+    int session;
+    int tty_nr;
+    int tpgid;
+    unsigned int flags;
+    unsigned long minflt, cminflt, majflt, cmajflt, utime, s_time;
+    long int cutime, cs_time, priority, niceval, num_threads, itrealvalue, vsize = 0, rss = 0, rsslim;
+    long long unsigned starttime;
 
+    pool = nx_pool_create_core();
+
+    if ( (rv = apr_file_open(&statfile, "/proc/self/stat", APR_READ, APR_OS_DEFAULT, pool)) == APR_SUCCESS )
+    {
+	nbytes = sizeof(statbuf);
+	memset(statbuf, 0, sizeof(statbuf));
+	if ( (rv = apr_file_read(statfile, statbuf, &nbytes)) == APR_SUCCESS )
+	{
+                 //                                            10                                      20
+	    if ( sscanf(statbuf, "%d %s %c %d %d %d %d %d %u %lu %lu %lu %lu %lu %lu %ld %ld %ld %ld %ld %ld %llu %lu %ld %lu",
+			&pid, comm, &state, &ppid, &pgrp, &session, &tty_nr, &tpgid, &flags, &minflt, 
+			&cminflt, &majflt, &cmajflt, &utime, &s_time, &cutime, &cs_time, &priority, &niceval, &num_threads,
+			&itrealvalue, &starttime, &vsize, &rss, &rsslim) > 0 )
+	    {
+		retval = (int) (rss * sysconf(_SC_PAGESIZE));
+	    }
+	}
+	
+	apr_file_close(statfile);
+    }
+    apr_pool_destroy(pool);
+
+    return ( retval );
+}
+*/
 
 void nx_module_shutdown_self(nx_module_t *module)
 {
     nx_module_status_t status;
+    int memsize;
+    int memsize2;
 
     ASSERT(module != NULL);
 
     log_debug("SHUTDOWN: %s", module->name);
+    //log_info("mem size: %d", get_mem_size());
 
     status = nx_module_get_status(module);
 
@@ -1332,10 +1391,17 @@ void nx_module_shutdown_self(nx_module_t *module)
     //FIXME: free module->vars
     //FIXME: free module->stats
 
+    //memsize = get_mem_size();
+    //log_info("mem size before pool destroy: %d", memsize);
+    
     if ( module->pool != NULL )
     {
 	apr_pool_destroy(module->pool);
     }
+
+    //memsize2 = get_mem_size();
+    //log_info("mem size after pool destroy: %d (freed: %d)", memsize2, memsize - memsize2);
+
 }
 
 
@@ -1388,6 +1454,7 @@ nx_expr_statement_list_t *nx_module_parse_exec_block(nx_module_t *module,
 	    }
 	    catch(e)
 	    {
+		log_debug("Error in Exec block: [%s]", curr->args);
 		rethrow_msg(e, "Couldn't parse Exec block at %s:%d",
 			    curr->filename, curr->line_num);
 	    }
@@ -1397,6 +1464,7 @@ nx_expr_statement_list_t *nx_module_parse_exec_block(nx_module_t *module,
 	    }
 	    else
 	    {
+		ASSERT(statements != NULL);
 		while ( (stmnt = NX_DLIST_FIRST(statements)) != NULL )
 		{
 		    NX_DLIST_REMOVE(statements, stmnt, link);
@@ -1414,9 +1482,10 @@ nx_expr_statement_list_t *nx_module_parse_exec_block(nx_module_t *module,
 
 static void nx_module_parse_schedule_blocks(nx_module_t *module)
 {
-    const nx_directive_t *curr, *curr2;
+    const nx_directive_t * volatile curr, * volatile curr2;
     nx_schedule_entry_t *sched;
     apr_time_t first;
+    nx_exception_t e;
 
     ASSERT(module->directives != NULL);
     curr = module->directives;
@@ -1451,7 +1520,14 @@ static void nx_module_parse_schedule_blocks(nx_module_t *module)
 			nx_conf_error(curr2, "'Every' is already defined within the Schedule block");
 		    }
 		    sched->type = NX_SCHEDULE_ENTRY_TYPE_CRONTAB;
-		    nx_schedule_entry_parse_crontab(sched, curr2->args);
+		    try
+		    {
+			nx_schedule_entry_parse_crontab(sched, curr2->args);
+		    }
+		    catch(e)
+		    {
+			nx_conf_error(curr2, "couldn't parse value for directive 'When': %s", nx_exception_get_message(&e, 0));
+		    }
 		}
 		else if ( strcasecmp(curr2->directive, "Every") == 0 )
 		{
@@ -1463,7 +1539,14 @@ static void nx_module_parse_schedule_blocks(nx_module_t *module)
 		    {
 			nx_conf_error(curr2, "'Every' is already defined within the Schedule block");
 		    }
-		    nx_schedule_entry_parse_every(sched, curr2->args);
+		    try
+		    {
+			nx_schedule_entry_parse_every(sched, curr2->args);
+		    }
+		    catch(e)
+		    {
+			nx_conf_error(curr2, "couldn't parse value for directive 'Every': %s", nx_exception_get_message(&e, 0));
+		    }
 		    sched->type = NX_SCHEDULE_ENTRY_TYPE_EVERY;
 		}
 		else if ( strcasecmp(curr2->directive, "First") == 0 )
@@ -1492,6 +1575,10 @@ static void nx_module_parse_schedule_blocks(nx_module_t *module)
 	    {
 		sched->first = first;
 	    }
+	    if ( sched->type == 0 )
+	    {
+		nx_conf_error(curr, "'When' or 'Every' must be defined in Schedule block");
+	    }
 
 	    NX_DLIST_INSERT_TAIL(module->schedule, sched, link);
 	}
@@ -1582,6 +1669,7 @@ const void *nx_module_get_resource(nx_module_t *module,
 nx_module_input_t *nx_module_input_new(nx_module_t *module, apr_pool_t *pool)
 {
     nx_module_input_t *input = NULL;
+    int bufsize = NX_MODULE_DEFAULT_INPUT_BUFSIZE;
 
     if ( pool == NULL )
     {
@@ -1589,15 +1677,19 @@ nx_module_input_t *nx_module_input_new(nx_module_t *module, apr_pool_t *pool)
     }
     input = apr_pcalloc(pool, sizeof(nx_module_input_t));
     input->pool = pool;
-    input->buf = apr_pcalloc(pool, NX_INPUT_BUFSIZE);
-    input->bufsize = NX_INPUT_BUFSIZE;
+    if ( (module->type == NX_MODULE_TYPE_INPUT) && (module->input.bufsize != 0) )
+    {
+	bufsize = module->input.bufsize;
+    }
+    input->buf = apr_pcalloc(pool, (apr_size_t) bufsize);
+    input->bufsize = bufsize;
     input->module = module;
 
     return ( input );
 }
 
 
-
+/*
 nx_module_output_t *nx_module_output_new(nx_module_t *module, apr_pool_t *pool)
 {
     nx_module_output_t *output = NULL;
@@ -1606,13 +1698,13 @@ nx_module_output_t *nx_module_output_new(nx_module_t *module, apr_pool_t *pool)
 
     output = apr_pcalloc(pool, sizeof(nx_module_output_t));
     output->pool = pool;
-    output->buf = apr_pcalloc(pool, NX_OUTPUT_BUFSIZE);
-    output->bufsize = NX_OUTPUT_BUFSIZE;
+    output->buf = apr_pcalloc(pool, NX_MODULE_DEFAULT_OUTPUT_BUFSIZE);
+    output->bufsize = NX_MODULE_DEFAULT_OUTPUT_BUFSIZE;
     output->module = module;
 
     return ( output );
 }
-
+*/
 
 
 void nx_module_input_free(nx_module_input_t *input)
@@ -1637,7 +1729,7 @@ void nx_module_input_free(nx_module_input_t *input)
 	    input->desc.f = NULL;
 	}
     }
-
+    ASSERT(input->pool != NULL);
     apr_pool_destroy(input->pool);
 }
 
@@ -1647,6 +1739,7 @@ void nx_module_output_free(nx_module_output_t *output)
 {
     ASSERT(output != NULL);
 
+    ASSERT(output->pool != NULL);
     apr_pool_destroy(output->pool);
 }
 
@@ -1723,20 +1816,29 @@ void *nx_module_output_data_get(nx_module_output_t *output, const char *key)
 }
 
 
+#define NX_MODULE_INPUT_NAME_LENGTH 250
 
 void nx_module_input_name_set(nx_module_input_t *input, const char *name)
 {
     ASSERT(input != NULL);
     ASSERT(name != NULL);
 
-    if ( input->name != NULL )
+    if ( input->name == NULL )
+    {
+	input->name = apr_pcalloc(input->pool, NX_MODULE_INPUT_NAME_LENGTH);
+	apr_cpystrn(input->name, name, NX_MODULE_INPUT_NAME_LENGTH);
+    }
+    else
     {
 	if ( strcmp(input->name, name) == 0 )
 	{ // already have the same, don't update
 	    return;
 	}
+	else
+	{
+	    apr_cpystrn(input->name, name, NX_MODULE_INPUT_NAME_LENGTH);
+	}
     }
-    input->name = apr_pstrdup(input->pool, name);
 }
 
 
@@ -1764,10 +1866,12 @@ const char *nx_module_input_name_get(nx_module_input_t *input)
 	{
 	    char *ipstr;
 
-	    apr_socket_addr_get(&sa, APR_REMOTE, input->desc.s);
-	    apr_sockaddr_ip_get(&ipstr, sa);
-	    input->name = apr_pstrdup(input->pool, ipstr);
-	    return ( input->name );
+	    if ( (apr_socket_addr_get(&sa, APR_REMOTE, input->desc.s) == APR_SUCCESS) &&
+		 (apr_sockaddr_ip_get(&ipstr, sa) == APR_SUCCESS) )
+	    {
+		input->name = apr_pstrdup(input->pool, ipstr);
+		return ( input->name );
+	    }
 	}
 	return ( "unknown" );
     }
@@ -1920,7 +2024,8 @@ nx_module_output_func_decl_t *nx_module_output_func_lookup(const char *fname)
 void nx_module_input_func_register(const nx_module_t *module,
 				   const char *fname,
 				   nx_module_input_func_t *func,
-				   const void *data)
+				   nx_module_input_func_t *flush,
+				   void *data)
 {
     nx_ctx_t *ctx;
     nx_module_input_func_decl_t *inputfunc;
@@ -1936,6 +2041,7 @@ void nx_module_input_func_register(const nx_module_t *module,
     ASSERT(inputfunc != NULL);
 
     inputfunc->func = func;
+    inputfunc->flush = flush;
     inputfunc->module = module;
     inputfunc->name = apr_pstrdup(ctx->pool, fname);
     inputfunc->data = data;
@@ -1949,7 +2055,7 @@ void nx_module_input_func_register(const nx_module_t *module,
 void nx_module_output_func_register(const nx_module_t *module,
 				    const char *fname,
 				    nx_module_output_func_t *func,
-				    const void *data)
+				    void *data)
 {
     nx_ctx_t *ctx;
     nx_module_output_func_decl_t *outputfunc;
@@ -2109,7 +2215,8 @@ int nx_module_parse_types(nx_value_type_t *types, char *string)
     return ( num_type );
 }
 
-
+// TODO: this size is pretty big and uses 500k per pollset
+#define NX_POLLSET_NUM	10000 // This also limits the number of connections per socket
 
 void nx_module_pollset_init(nx_module_t *module)
 {
@@ -2117,6 +2224,7 @@ void nx_module_pollset_init(nx_module_t *module)
     apr_status_t rv;
     int trycnt = 0;
     const char *methodname;
+    static apr_uint32_t _pollset_num = NX_POLLSET_NUM; // TODO: possible race condition with multiple modules
 
 #ifdef HAVE_APR_POLLSET_WAKEUP
     apr_uint32_t flags = APR_POLLSET_WAKEABLE;
@@ -2128,12 +2236,26 @@ void nx_module_pollset_init(nx_module_t *module)
     // windows 2003 with WSAEWOUDLBLOCK (10035) because of a non-blocking accept
     // http://osdir.com/ml/dev-apr-apache/2010-08/msg00039.html
     // So the following is a workaround
-    while ( APR_STATUS_IS_EAGAIN(rv = apr_pollset_create(&pollset, NX_POLLSET_NUM,
-							 module->pool, flags)) )
-    { 
-	apr_sleep(APR_USEC_PER_SEC / 10);
-	trycnt++;
-	if ( trycnt > 50 )
+    for ( ; ; )
+    {
+	rv = apr_pollset_create(&pollset, _pollset_num,
+				module->pool, flags);
+	if ( APR_STATUS_IS_EINVAL(rv) )
+	{
+	    // Detect the number of pollset entries this system can support
+	    // windows2003 gives an invalid argument above 500
+	    _pollset_num -= 10; // TODO: possible race condition with multiple modules
+	}
+	else if ( APR_STATUS_IS_EAGAIN(rv) )
+	{ 
+	    apr_sleep(APR_USEC_PER_SEC / 10);
+	    trycnt++;
+	    if ( trycnt > 50 )
+	    {
+		break;
+	    }
+	}
+	else
 	{
 	    break;
 	}
@@ -2183,9 +2305,14 @@ void nx_module_pollset_add_file(nx_module_t *module,
     ASSERT(module != NULL);
     ASSERT(file != NULL);
 
-    apr_file_data_get((void **) &setevents, "reqevents", file);
+    apr_file_data_get((void **) &setevents, NX_POLLSET_REQEVENTS_KEY, file);
     if ( (setevents != NULL) && (*setevents != 0) )
     {
+	if ( *setevents == reqevents )
+	{
+	    log_debug("file already added to pollset with reqevents [%x]", reqevents);
+	    return;
+	}
 	nx_module_pollset_remove_file(module, file);
     }
 
@@ -2216,7 +2343,7 @@ void nx_module_pollset_add_file(nx_module_t *module,
 	pool = apr_file_pool_get(file);
 	setevents = apr_palloc(pool, sizeof(apr_int16_t));
 	*setevents = reqevents;
-	apr_file_data_set(file, (void *) setevents, "reqevents", NULL);
+	apr_file_data_set(file, (void *) setevents, NX_POLLSET_REQEVENTS_KEY, NULL);
     }
     else
     {
@@ -2236,7 +2363,7 @@ void nx_module_pollset_remove_file(nx_module_t *module,
     ASSERT(module != NULL);
     ASSERT(file != NULL);
 
-    apr_file_data_get((void **) &reqevents, "reqevents", file);
+    apr_file_data_get((void **) &reqevents, NX_POLLSET_REQEVENTS_KEY, file);
     if ( reqevents == NULL )
     {
 	return;
@@ -2265,6 +2392,19 @@ void nx_module_pollset_remove_file(nx_module_t *module,
 }
 
 
+/*
+ * NOTE!!! We are using the apr_pollset without APR_POLLSET_NOCPOY.
+ * On Linux the epoll implementation allocates a copy of apr_pollfd_t out of the pool
+ * of the pollset. Here is how that works:
+ *  apr_pollset_remove() adds the copy of apr_pollfd_t to a Dead List.
+ *  apr_pollset_add() checks the Free List and uses an entry from there.
+ *  apr_pollset_poll adds all entries from the Dead List to the Free List
+ * Calling _add() + _remove() several times without a call to _poll() results in excess
+ * allocations from the pool of the pollset - i.e. leak.
+ *
+ * TODO: The code including all modules needs to be reviewed and checked whether
+ * it is possible for a module to emit multiple _add()s with different event requests.
+ */
 
 void nx_module_pollset_add_socket(nx_module_t *module,
 				  apr_socket_t *sock,
@@ -2279,11 +2419,18 @@ void nx_module_pollset_add_socket(nx_module_t *module,
     ASSERT(sock != NULL);
     ASSERT(module->pollset != NULL);
 
-    apr_socket_data_get((void **) &setevents, "reqevents", sock);
+    apr_socket_data_get((void **) &setevents, NX_POLLSET_REQEVENTS_KEY, sock);
     log_debug("add socket [%x]", reqevents);
     if ( (setevents != NULL) && (*setevents != 0) )
     {
-	log_debug("removing already added socket");
+	if ( *setevents == reqevents )
+	{
+	    log_debug("socket already added to pollset with reqevents [%x != %x]", *setevents, reqevents);
+	    return;
+	}
+	// FIXME: this can result in a leak due to the way apr_pollset works with
+	// APR_POLLSET_NOCPOY as described above
+	log_debug("removing already added socket with different reqevents");
 	nx_module_pollset_remove_socket(module, sock);
     }
     log_debug("socket added [%x]", reqevents);
@@ -2293,7 +2440,7 @@ void nx_module_pollset_add_socket(nx_module_t *module,
     pfd.reqevents = reqevents;
     pfd.client_data = module;
     pfd.p = module->pool;
-
+    
     if ( (rv = apr_pollset_add(module->pollset, &pfd)) != APR_SUCCESS )
     {
 	char errmsg[200];
@@ -2315,7 +2462,7 @@ void nx_module_pollset_add_socket(nx_module_t *module,
 	pool = apr_socket_pool_get(sock);
 	setevents = apr_palloc(pool, sizeof(apr_int16_t));
 	*setevents = reqevents;
-	apr_socket_data_set(sock, (void *) setevents, "reqevents", NULL);
+	apr_socket_data_set(sock, (void *) setevents, NX_POLLSET_REQEVENTS_KEY, NULL);
     }
     else
     {
@@ -2335,9 +2482,10 @@ void nx_module_pollset_remove_socket(nx_module_t *module,
     ASSERT(module != NULL);
     ASSERT(sock != NULL);
 
-    apr_socket_data_get((void **) &reqevents, "reqevents", sock);
+    apr_socket_data_get((void **) &reqevents, NX_POLLSET_REQEVENTS_KEY, sock);
     if ( reqevents == NULL )
     {
+	//log_debug("nothing to remove in nx_module_pollset_remove_socket()");
 	return;
     }
     log_debug("remove socket [%x]", *reqevents);
@@ -2484,6 +2632,13 @@ void nx_module_pollset_poll(nx_module_t *module, boolean readd)
 		eventcnt++;
 	    }
 
+	    if ( (ret_pfd[i].rtnevents & APR_POLLNVAL) != 0 )
+	    { // Descriptor invalid 
+		log_error("invalid descriptor in poll, removing");
+		apr_pollset_remove(module->pollset, &ret_pfd[i]);
+		apr_sleep(APR_USEC_PER_SEC * 1); //sleep 1 sec in order to avoid flooding logs if it breaks totally
+	    }
+
 	    if ( ret_pfd[i].rtnevents == APR_POLLERR )
 	    {
 		log_error("Module %s got POLLERR", module->name);
@@ -2494,6 +2649,7 @@ void nx_module_pollset_poll(nx_module_t *module, boolean readd)
 		{
 		    log_error("unhandled event! (reqevents: %d, rtnevents: %d)", 
 			      ret_pfd[i].reqevents, ret_pfd[i].rtnevents);
+		    apr_sleep(APR_USEC_PER_SEC * 1); //sleep 1 sec in order to avoid flooding logs if it breaks totally
 		}
 	    }
 
