@@ -6,6 +6,7 @@
  */
 
 #include <apr_lib.h>
+#include <stdlib.h>
 
 #include "error_debug.h"
 #include "exception.h"
@@ -13,6 +14,15 @@
 
 #define NX_LOGMODULE NX_LOGMODULE_CORE
 
+static uint32_t _string_limit = NX_STRING_DEFAULT_LIMIT;
+
+
+uint32_t nx_string_get_limit()
+{
+    return ( _string_limit );
+}
+
+
 
 /*
  * Frees only dynamic part of the string, not itself
@@ -72,11 +82,12 @@ nx_string_t *nx_string_new_size(size_t len)
 	return ( nx_string_new() );
     }
 
-    retval = malloc(sizeof(nx_string_t));
-    if ( len > NX_STRING_MAX_SIZE )
+    if ( len > _string_limit )
     {
-	throw_msg("oversized string, limit is %d bytes", NX_STRING_MAX_SIZE);
+	throw_msg("oversized string, limit is %d bytes", _string_limit);
     }
+
+    retval = malloc(sizeof(nx_string_t));
     retval->buf = malloc(len);
     *(retval->buf) = '\0';
     retval->bufsize = (uint32_t) len;
@@ -98,9 +109,9 @@ nx_string_t *nx_string_create(const char *src, int len)
     {
 	len = (int) strlen(src);
     }
-    if ( len > NX_STRING_MAX_SIZE )
+    if ( len > (int) _string_limit )
     {
-	throw_msg("oversized string, limit is %d bytes", NX_STRING_MAX_SIZE);
+	throw_msg("oversized string, limit is %d bytes", _string_limit);
     }
 
     retval = malloc(sizeof(nx_string_t));
@@ -116,6 +127,32 @@ nx_string_t *nx_string_create(const char *src, int len)
 
 
 
+nx_string_t *nx_string_create_owned(char *src, int len)
+{
+    nx_string_t *retval;
+
+    ASSERT(src != NULL);
+
+    if ( len < 0 )
+    {
+	len = (int) strlen(src);
+    }
+    if ( len > (int) _string_limit )
+    {
+	throw_msg("oversized string, limit is %d bytes", _string_limit);
+    }
+
+    retval = malloc(sizeof(nx_string_t));
+    retval->buf = src;
+    retval->bufsize = (uint32_t) len + 1;
+    retval->len = (uint32_t) len;
+    retval->flags = 0;
+
+    return ( retval );
+}
+
+
+
 nx_string_t *nx_string_init_const(nx_string_t *dst, const char *src)
 {
     ASSERT(dst != NULL);
@@ -140,9 +177,9 @@ static void ensure_size(nx_string_t *dst, size_t len)
 	return;
     }
 
-    if ( len > NX_STRING_MAX_SIZE )
+    if ( len > _string_limit )
     {
-	throw_msg("string limit (%d bytes) reached", NX_STRING_MAX_SIZE);
+	throw_msg("string limit (%d bytes) reached", _string_limit);
     }
 
     if ( dst->bufsize < 1024 )
@@ -152,9 +189,9 @@ static void ensure_size(nx_string_t *dst, size_t len)
     else
     {
 	newsize = (dst->bufsize * 3) / 2;
-	if ( newsize > NX_STRING_MAX_SIZE )
+	if ( newsize > _string_limit )
 	{
-	    newsize = NX_STRING_MAX_SIZE;
+	    newsize = _string_limit;
 	}
     }
     if ( newsize <= len )
@@ -217,10 +254,28 @@ nx_string_t *nx_string_append(nx_string_t *dst, const char *src, int srclen)
 	return ( dst );
     }
 
-    ensure_size(dst, dst->len + (size_t) srclen + 1);
-    memcpy(dst->buf + dst->len, src, (apr_size_t) srclen);
-    dst->len += (uint32_t) srclen;
-    dst->buf[dst->len] = '\0';
+    if ( dst->len + (size_t) srclen + 1 > _string_limit )
+    {
+	uint32_t len;
+
+	len = _string_limit - 1;
+	if ( dst->len < len )
+	{
+	    ensure_size(dst, len + 1);
+	    memcpy(dst->buf + dst->len, src, (apr_size_t) (len - dst->len));
+	    dst->len = len;
+	    dst->buf[dst->len] = '\0';
+	    log_warn("string limit (%u) exceeded while trying to append", _string_limit);
+	}
+	// else silently discard since we have probably warned already
+    }
+    else
+    {
+	ensure_size(dst, dst->len + (size_t) srclen + 1);
+        memcpy(dst->buf + dst->len, src, (apr_size_t) srclen);
+	dst->len += (uint32_t) srclen;
+	dst->buf[dst->len] = '\0';
+    }
 
     return ( dst );
 }
@@ -241,12 +296,38 @@ nx_string_t *nx_string_prepend(nx_string_t *dst, const char *src, int srclen)
     {
 	return ( dst );
     }
-    ensure_size(dst, dst->len + (size_t) srclen + 1);
-    memmove(dst->buf + srclen, dst->buf, (size_t) dst->len);
-    memcpy(dst->buf, src, (size_t) srclen);
-    dst->len += (uint32_t) srclen;
-    dst->buf[dst->len] = '\0';
 
+    if ( dst->len + (size_t) srclen + 1 > _string_limit )
+    {
+	uint32_t len;
+
+	len = _string_limit - 1;
+	if ( dst->len < len )
+	{
+	    ensure_size(dst, len + 1);
+	    if ( srclen > _string_limit )
+	    {
+		memcpy(dst->buf, src, (size_t) len);
+	    }
+	    else
+	    {
+		memmove(dst->buf + srclen, dst->buf, (size_t) len - srclen);
+		memcpy(dst->buf, src, (size_t) srclen);
+	    }
+	    dst->len = len;
+	    dst->buf[dst->len] = '\0';
+	    log_warn("string limit (%u) exceeded while trying to preppend", _string_limit);
+	}
+	// else silently discard since we have probably warned already
+    }
+    else
+    {
+	ensure_size(dst, dst->len + (size_t) srclen + 1);
+	memmove(dst->buf + srclen, dst->buf, (size_t) dst->len);
+	memcpy(dst->buf, src, (size_t) srclen);
+	dst->len += (uint32_t) srclen;
+	dst->buf[dst->len] = '\0';
+    }
     return ( dst );
 }
 
@@ -279,15 +360,22 @@ nx_string_t *nx_string_set(nx_string_t *dst, const char *src, int srclen)
 nx_string_t *nx_string_clone(const nx_string_t *str)
 {
     nx_string_t *retval;
+    uint32_t len;
 
     ASSERT(str != NULL);
     ASSERT(str->buf != NULL);
-
-    retval = nx_string_new_size(str->len + 1);
+    
+    len = str->len;
+    if ( len > _string_limit )
+    {
+	log_warn("truncating oversized string (%u) to StringLimit (%u) in nx_string_clone()", len, _string_limit);
+	len = _string_limit - 1;
+    }
+    retval = nx_string_new_size(len + 1);
     retval->flags = 0;
-    retval->len = str->len;
-    memcpy(retval->buf, str->buf, str->len);
-    retval->buf[str->len] = '\0';
+    retval->len = len;
+    memcpy(retval->buf, str->buf, len);
+    retval->buf[len] = '\0';
 
     return ( retval );
 }
@@ -337,8 +425,17 @@ nx_string_t *nx_string_sprintf_append(nx_string_t 	*str,
     len = apr_vsnprintf(NULL, 0, fmt, ap);
     va_end(ap);
 
-    ensure_size(str, str->len + (size_t) len + 1);
+    if ( str->len + (size_t) len + 1 > _string_limit )
+    {
+	len = _string_limit - 1;
+	if ( str->len < len )
+	{
+	    log_warn("string limit (%u) exceeded while trying to append", _string_limit);
+	}
+    }
 
+    ensure_size(str, str->len + (size_t) len + 1);
+    
     va_start(ap, fmt);
     ASSERT(apr_vsnprintf(str->buf + str->len, (apr_size_t) len + 1, fmt, ap) == len);
     va_end(ap);
@@ -527,8 +624,8 @@ size_t nx_string_unescape_c(char *str)
  * Finds the next UTF-8 character in the string after p.
  */
 
-static char *nx_utf8_find_next_char(char *p,
-				    char *end)
+char *nx_utf8_find_next_char(char *p,
+			     char *end)
 {
     ASSERT(p != NULL);
 
@@ -552,8 +649,8 @@ static char *nx_utf8_find_next_char(char *p,
  * Check if a utf-8 character is valid
  */
 
-static boolean nx_utf8_is_valid_char(const char *src,
-				     int32_t length)
+boolean nx_utf8_is_valid_char(const char *src,
+			      int32_t length)
 {
     char a;
     const char *srcptr = src + length;
