@@ -75,10 +75,15 @@ void nx_value_kill(nx_value_t *value)
 		    free(value->regexp.str);
 		    value->regexp.str = NULL;
 		}
+		if ( value->regexp.replacement != NULL )
+		{
+		    free(value->regexp.replacement);
+		}
 		break;    
 	    default:
 		nx_panic("invalid value type: %d", value->type);
 	}
+	value->defined = FALSE; // make sure we will not be called twice
     }
 }
 
@@ -108,8 +113,6 @@ nx_value_t *nx_value_clone(nx_value_t *dst, const nx_value_t *value)
 	ASSERT(retval != NULL);
     }
 
-    retval->defined = value->defined;
-    retval->type = value->type;
     if ( value->defined == TRUE )
     {
 	switch ( value->type )
@@ -160,6 +163,9 @@ nx_value_t *nx_value_clone(nx_value_t *dst, const nx_value_t *value)
 		nx_panic("invalid value type: %d", value->type);
 	}
     }
+    retval->defined = value->defined;
+    retval->type = value->type;
+
     return ( retval );
 }
 
@@ -204,7 +210,7 @@ nx_value_t *nx_value_new_regexp(const char *str)
     const char *error = NULL;
     int erroroffs = 0;
     int rc;
-    int size;
+    size_t size;
 
     ASSERT(str != NULL);
 
@@ -314,6 +320,7 @@ nx_value_t *nx_value_new(nx_value_type_t type)
     nx_value_t *retval;
 
     retval = malloc(sizeof(nx_value_t));
+    memset(retval, 0, sizeof(nx_value_t));
     ASSERT(retval != NULL);
     retval->type = type;
     retval->defined = TRUE;
@@ -579,8 +586,12 @@ int64_t nx_value_parse_int(const char *string)
 	    case 'G':
 		mul = 1024 * 1024 * 1024;
 		break;
+	    case '\r':
+		throw_msg("cannot parse integer \"%s\", invalid modifier: '\\r'", string);
+	    case '\n':
+		throw_msg("cannot parse integer \"%s\", invalid modifier: '\\n'", string);
 	    default:
-		throw_msg("invalid modifier: '%c'", string[i + len - 1]);
+		throw_msg("cannot parse integer \"%s\", invalid modifier: '%c'", string, string[i + len - 1]);
 	}
 	len--;
     }
@@ -609,9 +620,8 @@ int64_t nx_value_parse_int(const char *string)
 
 
 
-nx_value_t *nx_value_parse_ip4addr(const char *string)
+nx_value_t *nx_value_parse_ip4addr(nx_value_t *retval, const char *string)
 {
-    nx_value_t *retval = NULL;
     int ip4addr[4] = { 0, 0, 0, 0 };
     int i = 0;
     const char *ptr;
@@ -656,7 +666,10 @@ nx_value_t *nx_value_parse_ip4addr(const char *string)
 	throw_msg("invalid ipv4 address: %s", string);
     }
 
-    retval = nx_value_new(NX_VALUE_TYPE_IP4ADDR);
+    if ( retval == NULL)
+    {
+	retval = nx_value_new(NX_VALUE_TYPE_IP4ADDR);
+    }
     retval->ip4addr[0] = (uint8_t) ip4addr[0];
     retval->ip4addr[1] = (uint8_t) ip4addr[1];
     retval->ip4addr[2] = (uint8_t) ip4addr[2];
@@ -723,7 +736,7 @@ nx_value_t *nx_value_from_string(const char *string, nx_value_type_t type)
 	case NX_VALUE_TYPE_DATETIME:
 	    if ( nx_date_parse(&t, string, NULL) != APR_SUCCESS )
 	    {
-		return ( NULL );
+		throw_msg("Couldn't parse datetime value: '%s'", string);
 	    }
 	    retval = nx_value_new(type);
 	    retval->datetime = t;
@@ -748,7 +761,7 @@ nx_value_t *nx_value_from_string(const char *string, nx_value_type_t type)
 	    retval = nx_value_new_regexp(string);
 	    break;
 	case NX_VALUE_TYPE_IP4ADDR:
-	    retval = nx_value_parse_ip4addr(string);
+	    retval = nx_value_parse_ip4addr(NULL, string);
 	    break;
 	case NX_VALUE_TYPE_IP6ADDR:
 	    retval = nx_value_parse_ip6addr(string);
