@@ -12,8 +12,6 @@
 #include "../../../common/logdata.h"
 #include "../../../common/module.h"
 
-#define NX_CSV_MAX_FIELDS 150
-
 typedef enum nx_csv_quote_method_t
 {
     NX_CSV_QUOTE_METHOD_STRING = 1,
@@ -32,6 +30,7 @@ typedef struct nx_csv_ctx_t
     int num_type;
     const char *fields[NX_MODULE_MAX_FIELDS];
     nx_value_type_t types[NX_MODULE_MAX_FIELDS];
+    const char *undefvalue;
 } nx_csv_ctx_t;
 
 void nx_csv_ctx_set_quotechar(nx_csv_ctx_t *ctx, const char quotechar);
