@@ -78,11 +78,11 @@ static boolean _reopen_logfile(const char *filename)
 
 void nx_expr_proc__xm_fileop_file_cycle(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
-    nx_expr_arg_t *val;
+    nx_expr_list_elem_t *val;
     nx_value_t value;
     apr_status_t rv;
     apr_pool_t *pool = NULL;
@@ -111,7 +111,6 @@ void nx_expr_proc__xm_fileop_file_cycle(nx_expr_eval_ctx_t *eval_ctx,
 	throw_msg("string type required for 'file'");
     }
 
-
     val = NX_DLIST_NEXT(arg, link);
     if ( val != NULL )
     {
@@ -216,9 +215,9 @@ void nx_expr_proc__xm_fileop_file_cycle(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_rename(nx_expr_eval_ctx_t *eval_ctx,
 					 nx_module_t *module,
-					 nx_expr_arg_list_t *args)
+					 nx_expr_list_t *args)
 {
-    nx_expr_arg_t *src, *dst;
+    nx_expr_list_elem_t *src, *dst;
     nx_value_t srcval, dstval;
     nx_exception_t e;
     apr_status_t rv;
@@ -282,9 +281,9 @@ void nx_expr_proc__xm_fileop_file_rename(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_copy(nx_expr_eval_ctx_t *eval_ctx,
 				       nx_module_t *module,
-				       nx_expr_arg_list_t *args)
+				       nx_expr_list_t *args)
 {
-    nx_expr_arg_t *src, *dst;
+    nx_expr_list_elem_t *src, *dst;
     nx_value_t srcval, dstval;
     nx_exception_t e;
     apr_status_t rv;
@@ -349,9 +348,9 @@ void nx_expr_proc__xm_fileop_file_copy(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_remove(nx_expr_eval_ctx_t *eval_ctx,
 					 nx_module_t *module,
-					 nx_expr_arg_list_t *args)
+					 nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
     apr_status_t rv;
     apr_pool_t *pool = NULL;
@@ -363,7 +362,7 @@ void nx_expr_proc__xm_fileop_file_remove(nx_expr_eval_ctx_t *eval_ctx,
     char *idx;
     int flags = 0;
     apr_finfo_t finfo;
-    nx_expr_arg_t *older;
+    nx_expr_list_elem_t *older;
     nx_value_t olderval;
     apr_time_t older_time = 0LL;
 
@@ -409,8 +408,10 @@ void nx_expr_proc__xm_fileop_file_remove(nx_expr_eval_ctx_t *eval_ctx,
 	    pool = nx_pool_create_core();
 
 	    filename = file.string->buf;
+	    log_debug("file_remove() called with wildcarded path: %s", filename);
+
 	    idx = strrchr(filename, '/');
-#ifndef WIN32
+#ifdef WIN32
 	    flags = APR_FNM_CASE_BLIND;
 	    if ( idx == NULL ) 
 	    {
@@ -426,6 +427,7 @@ void nx_expr_proc__xm_fileop_file_remove(nx_expr_eval_ctx_t *eval_ctx,
 		dirname = nx_string_create(filename, (int) (idx + 1 - filename));
 		filename = idx + 1;
 	    }
+	    log_debug("file_remove(): checking for matching files under %s", dirname->buf);
 
 	    CHECKERR_MSG(apr_dir_open(&dir, dirname->buf, pool),
 			 "failed to open directory: %s", dirname->buf);
@@ -464,37 +466,49 @@ void nx_expr_proc__xm_fileop_file_remove(nx_expr_eval_ctx_t *eval_ctx,
 		}
 	    }
 	    nx_string_free(fname);
+	    fname = NULL;
 	    apr_pool_destroy(pool);
+	    pool = NULL;
 	}
 	else
 	{
-	    CHECKERR_MSG(apr_file_remove(file.string->buf, NULL), 
-			 "failed to remove file '%s'", file.string->buf);
+	    rv = apr_file_remove(file.string->buf, NULL);
+	    if ( APR_STATUS_IS_ENOENT(rv) )
+	    {
+	    }
+	    else
+	    {
+		CHECKERR_MSG(rv, "failed to remove file '%s'", file.string->buf);
+	    }
 	}
     }
     catch(e)
     {
-	nx_value_kill(&file);
-	if ( pool != NULL )
-	{
-	    apr_pool_destroy(pool);
-	}
-	if ( dirname != NULL )
-	{
-	    nx_string_free(dirname);
-	}
 	log_exception(e);
     }
+    nx_value_kill(&file);
+    if ( pool != NULL )
+    {
+	apr_pool_destroy(pool);
+    }
+    if ( dirname != NULL )
+    {
+	nx_string_free(dirname);
+    }
+    if ( fname != NULL )
+    {
+	nx_string_free(fname);
+    }
 }
 
 
 
 void nx_expr_proc__xm_fileop_file_link(nx_expr_eval_ctx_t *eval_ctx,
 				       nx_module_t *module,
-				       nx_expr_arg_list_t *args)
+				       nx_expr_list_t *args)
 {
 #ifdef HAVE_APR_FILE_LINK
-    nx_expr_arg_t *src, *dst;
+    nx_expr_list_elem_t *src, *dst;
     nx_value_t srcval, dstval;
     nx_exception_t e;
     apr_status_t rv;
@@ -558,9 +572,9 @@ void nx_expr_proc__xm_fileop_file_link(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_append(nx_expr_eval_ctx_t *eval_ctx,
 					 nx_module_t *module,
-					 nx_expr_arg_list_t *args)
+					 nx_expr_list_t *args)
 {
-    nx_expr_arg_t *src, *dst;
+    nx_expr_list_elem_t *src, *dst;
     nx_value_t srcval, dstval;
     nx_exception_t e;
     apr_status_t rv;
@@ -625,11 +639,11 @@ void nx_expr_proc__xm_fileop_file_append(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_write(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
-    nx_expr_arg_t *val;
+    nx_expr_list_elem_t *val;
     nx_value_t value;
     nx_exception_t e;
     apr_status_t rv;
@@ -705,11 +719,11 @@ void nx_expr_proc__xm_fileop_file_write(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_truncate(nx_expr_eval_ctx_t *eval_ctx,
 					   nx_module_t *module,
-					   nx_expr_arg_list_t *args)
+					   nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
-    nx_expr_arg_t *offset;
+    nx_expr_list_elem_t *offset;
     nx_value_t offsetval;
     nx_exception_t e;
     apr_off_t offs = 0;
@@ -788,11 +802,11 @@ void nx_expr_proc__xm_fileop_file_truncate(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_chown(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
-    nx_expr_arg_t *uid, *gid;
+    nx_expr_list_elem_t *uid, *gid;
     nx_value_t uidval, gidval;
     nx_exception_t e;
     int rv;
@@ -808,7 +822,7 @@ void nx_expr_proc__xm_fileop_file_chown(nx_expr_eval_ctx_t *eval_ctx,
     ASSERT(uid != NULL);
     ASSERT(uid->expr != NULL);
 
-    gid = NX_DLIST_NEXT(arg, link);
+    gid = NX_DLIST_NEXT(uid, link);
     ASSERT(gid != NULL);
     ASSERT(gid->expr != NULL);
 
@@ -868,23 +882,138 @@ void nx_expr_proc__xm_fileop_file_chown(nx_expr_eval_ctx_t *eval_ctx,
 #ifdef HAVE_CHOWN
     if ( (rv = chown(file.string->buf, (uid_t) uidval.integer, (gid_t) gidval.integer)) != 0 )
     {
-	log_aprerror(rv, "failed to change file ownership on '%s'", file.string->buf);
+	log_errno("failed to change file ownership on '%s'", file.string->buf);
     }
 #else
     log_error("This platform does not support the file_chown() function");
 #endif    
     nx_value_kill(&file);
+    nx_value_kill(&uidval);
+    nx_value_kill(&gidval);
+}
+
+
+
+void nx_expr_proc__xm_fileop_file_chown_usr_grp(nx_expr_eval_ctx_t *eval_ctx,
+					nx_module_t *module,
+					nx_expr_list_t *args)
+{
+    nx_expr_list_elem_t *arg;
+    nx_value_t file;
+    nx_expr_list_elem_t *user, *group;
+    nx_value_t usrval, grpval;
+    nx_exception_t e;
+    int rv;
+    apr_status_t rv2;
+    apr_uid_t uid;
+    apr_gid_t tmp;
+    apr_gid_t gid;
+
+    ASSERT(module != NULL);
+
+    ASSERT(args != NULL);
+    arg = NX_DLIST_FIRST(args);
+    ASSERT(arg != NULL);
+    ASSERT(arg->expr != NULL);
+
+    user = NX_DLIST_NEXT(arg, link);
+    ASSERT(user != NULL);
+    ASSERT(user->expr != NULL);
+
+    group = NX_DLIST_NEXT(user, link);
+    ASSERT(group != NULL);
+    ASSERT(group->expr != NULL);
+
+    nx_expr_evaluate(eval_ctx, &file, arg->expr);
+
+    if ( file.defined != TRUE )
+    {
+	throw_msg("'file' is undef");
+    }
+    if ( file.type != NX_VALUE_TYPE_STRING )
+    {
+	nx_value_kill(&file);
+	throw_msg("string type required for 'file'");
+    }
+
+    try
+    {
+	nx_expr_evaluate(eval_ctx, &usrval, user->expr);
+    }
+    catch(e)
+    {
+	nx_value_kill(&file);
+	rethrow(e);
+    }
+    if ( usrval.defined != TRUE )
+    {
+	nx_value_kill(&file);
+	throw_msg("'user' is undef");
+    }
+    if ( usrval.type != NX_VALUE_TYPE_STRING )
+    {
+	nx_value_kill(&file);
+	throw_msg("string type required for 'user'");
+    }
+
+    try
+    {
+	nx_expr_evaluate(eval_ctx, &grpval, group->expr);
+    }
+    catch(e)
+    {
+	nx_value_kill(&file);
+	rethrow(e);
+    }
+    if ( grpval.defined != TRUE )
+    {
+	nx_value_kill(&file);
+	throw_msg("'group' is undef");
+    }
+    if ( grpval.type != NX_VALUE_TYPE_STRING )
+    {
+	nx_value_kill(&file);
+	throw_msg("string type required for 'group'");
+    }
+    
+    if ( (rv2 = apr_gid_get(&gid, grpval.string->buf, module->pool)) != APR_SUCCESS )
+    {
+	log_aprerror(rv2, "failed to change file ownership on '%s', couldn't resolve group name '%s'",
+		     file.string->buf, grpval.string->buf);
+    }
+    else
+    { // APR_SUCCESS
+	if ( (rv2 = apr_uid_get(&uid, &tmp, usrval.string->buf, module->pool)) != APR_SUCCESS )
+	{
+	    log_aprerror(rv2, "failed to change file ownership on '%s', couldn't resolve user name '%s'",
+			 file.string->buf, usrval.string->buf);
+	}
+	else
+	{ // APR_SUCCESS
+#ifdef HAVE_CHOWN
+	    if ( (rv = chown(file.string->buf, (uid_t)uid, (gid_t)gid)) != 0 )
+	    {
+		log_errno("failed to change file ownership on '%s'", file.string->buf);
+	    }
+#else
+	    log_error("This platform does not support the file_chown() function");
+#endif    
+	}
+    }
+    nx_value_kill(&file);
+    nx_value_kill(&usrval);
+    nx_value_kill(&grpval);
 }
 
 
 
 void nx_expr_proc__xm_fileop_file_chmod(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
-    nx_expr_arg_t *mode;
+    nx_expr_list_elem_t *mode;
     nx_value_t modeval;
     nx_exception_t e;
     int rv;
@@ -947,9 +1076,9 @@ void nx_expr_proc__xm_fileop_file_chmod(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_file_touch(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t file;
     apr_status_t rv;
     apr_pool_t *pool;
@@ -1007,9 +1136,9 @@ void nx_expr_proc__xm_fileop_file_touch(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_dir_make(nx_expr_eval_ctx_t *eval_ctx,
 				      nx_module_t *module,
-				      nx_expr_arg_list_t *args)
+				      nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t path;
     apr_status_t rv;
     apr_pool_t *pool;
@@ -1046,9 +1175,9 @@ void nx_expr_proc__xm_fileop_dir_make(nx_expr_eval_ctx_t *eval_ctx,
 
 void nx_expr_proc__xm_fileop_dir_remove(nx_expr_eval_ctx_t *eval_ctx,
 					nx_module_t *module,
-					nx_expr_arg_list_t *args)
+					nx_expr_list_t *args)
 {
-    nx_expr_arg_t *arg;
+    nx_expr_list_elem_t *arg;
     nx_value_t path;
     apr_status_t rv;
     apr_pool_t *pool;
@@ -1277,7 +1406,7 @@ void nx_expr_func__xm_fileop_file_dirname(nx_expr_eval_ctx_t *eval_ctx UNUSED,
 
     filename = args[0].string->buf;
     idx = strrchr(filename, '/');
-#ifndef WIN32
+#ifdef WIN32
     if ( idx == NULL ) 
     {
         idx = strrchr(filename, '\\');
