@@ -409,6 +409,11 @@ static void xm_multiline_config(nx_module_t *module)
 		{
 		    modconf->headerline_expr = nx_expr_parse(module, curr->args, module->pool, curr->filename,
 							     curr->line_num, curr->argsstart);
+		    if ( modconf->headerline_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for HeaderLine: '%s'", curr->args);
+		    }
+
 		    if ( !((modconf->headerline_expr->rettype == NX_VALUE_TYPE_STRING) ||
 			   (modconf->headerline_expr->rettype == NX_VALUE_TYPE_REGEXP)) )
 		    {
@@ -451,6 +456,10 @@ static void xm_multiline_config(nx_module_t *module)
 		{
 		    modconf->endline_expr = nx_expr_parse(module, curr->args, module->pool, curr->filename,
 							     curr->line_num, curr->argsstart);
+		    if ( modconf->endline_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for EndLine: '%s'", curr->args);
+		    }
 		    if ( !((modconf->endline_expr->rettype == NX_VALUE_TYPE_STRING) ||
 			   (modconf->endline_expr->rettype == NX_VALUE_TYPE_REGEXP)) )
 		    {
