@@ -407,7 +407,7 @@ static void set_syslog_message(nx_logdata_t *logdata,
 #define IS_HOSTCHAR(c) ( ((c >= '0') && (c <= '9')) || ((c >= 'a') && (c <= 'z')) || \
   ((c >= 'A') && (c <= 'Z')) || (c == '.') || (c == '_') || (c == '-') || (c == '/') )
 #define IS_TAGCHAR(c) ( ((c >= '0') && (c <= '9')) || ((c >= 'a') && (c <= 'z')) || \
-  ((c >= 'A') && (c <= 'Z')) || (c == '_') || (c == '-') || (c == '/') )
+  ((c >= 'A') && (c <= 'Z')) || (c == '_') || (c == '-') || (c == '/') || (c == '.') )
 
 boolean nx_syslog_parse_rfc3164(nx_logdata_t *logdata,
 				const char *string,
@@ -683,18 +683,31 @@ void nx_logdata_to_syslog_rfc3164(nx_logdata_t *logdata)
     nx_value_t timestamp;
     nx_value_t application;
     nx_value_t pid;
-    nx_value_t message;
+    nx_value_t msg;
+    nx_string_t *tmpmsg = NULL;
     size_t len;
     char tmpstr[20];
     int i;
 
     ASSERT(logdata != NULL);
     ASSERT(logdata->raw_event != NULL);
-    
+
+    if ( (nx_logdata_get_field_value(logdata, "Message", &msg) == TRUE) &&
+	 (msg.type == NX_VALUE_TYPE_STRING) && (msg.defined == TRUE) )
+    {
+	// we have a Message field
+    }
+    else
+    { // otherwise use raw_event
+	tmpmsg = nx_string_clone(logdata->raw_event);
+	msg.string = tmpmsg;
+	msg.type = NX_VALUE_TYPE_STRING;
+    }
+
     pri = nx_syslog_get_priority(logdata);
 
     if ( (nx_logdata_get_field_value(logdata, "EventTime", &timestamp) == TRUE) &&
-	 (timestamp.type = NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
+	 (timestamp.type == NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
     {
 	nx_date_to_rfc3164(tmpstr, sizeof(tmpstr), timestamp.datetime);
     }
@@ -751,20 +764,22 @@ void nx_logdata_to_syslog_rfc3164(nx_logdata_t *logdata)
 	nx_string_append(logdata->raw_event, ":", 1);
     }
 
-    if ( (nx_logdata_get_field_value(logdata, "Message", &message) == TRUE) &&
-	 (message.type == NX_VALUE_TYPE_STRING) && (message.defined == TRUE) )
-    {
-	i = (int) logdata->raw_event->len;
-	nx_string_append(logdata->raw_event, " ", 1);
-	nx_string_append(logdata->raw_event, message.string->buf, (int) message.string->len);
-	for ( ; i < (int) logdata->raw_event->len; i++ )
-	{ // replace linebreaks with space
-	    if ( (logdata->raw_event->buf[i] == '\n') || (logdata->raw_event->buf[i] == '\r') )
-	    {
-		logdata->raw_event->buf[i] = ' ';
-	    }
+    // Append message
+    i = (int) logdata->raw_event->len;
+    nx_string_append(logdata->raw_event, " ", 1);
+    nx_string_append(logdata->raw_event, msg.string->buf, (int) msg.string->len);
+    for ( ; i < (int) logdata->raw_event->len; i++ )
+    { // replace linebreaks with space
+	if ( (logdata->raw_event->buf[i] == '\n') || (logdata->raw_event->buf[i] == '\r') )
+	{
+	    logdata->raw_event->buf[i] = ' ';
 	}
     }
+
+    if (tmpmsg != NULL)
+    { // clean up temp copy
+	nx_string_free(tmpmsg);
+    }
 }
 
 
@@ -1076,7 +1091,7 @@ boolean nx_syslog_parse_rfc5424(nx_logdata_t *logdata,
     if ( *ptr == ' ' ) ptr++; // skip space
 
     // MESSAGE
-    if ( (ptr[0] == 0xEF) && (ptr[0] == 0xBB) && (ptr[0] == 0xBF) )
+    if ( (ptr[0] == 0xEF) && (ptr[1] == 0xBB) && (ptr[2] == 0xBF) )
     { //Skip UTF8 BOM
 	ptr += 3;
     }
@@ -1095,6 +1110,7 @@ static void nx_syslog_add_structured_data(nx_logdata_t *logdata)
     int cnt = 0;
     char *value = NULL;
     int i;
+    uint32_t sdlen;
     static const char *ignorelist[] = 
     {
 	"raw_event",
@@ -1138,7 +1154,17 @@ static void nx_syslog_add_structured_data(nx_logdata_t *logdata)
 		nx_string_append(sd, " ", 1);
 	    }
 	    cnt++;
+	    
+	    sdlen = sd->len;
 	    nx_string_append(sd, field->key, -1);
+	    for ( ;  sdlen < sd->len; sdlen++ )
+	    { // replace space, ], " with underscore in field names
+		if ( (sd->buf[sdlen] == ' ') || (sd->buf[sdlen] == ']') || (sd->buf[sdlen] == '"') )
+		{
+		    sd->buf[sdlen] = '_';
+		}
+	    }
+
 	    nx_string_append(sd, "=\"", 2);
 	    if ( field->value->type == NX_VALUE_TYPE_STRING )
 	    {
@@ -1190,7 +1216,8 @@ void nx_logdata_to_syslog_rfc5424(nx_logdata_t *logdata, boolean gmt)
     nx_value_t application;
     nx_value_t messageid;
     nx_value_t pid;
-    nx_value_t message;
+    nx_value_t msg;
+    nx_string_t *tmpmsg = NULL;
     size_t len;
     char tmpstr[33];
     int i;
@@ -1198,10 +1225,22 @@ void nx_logdata_to_syslog_rfc5424(nx_logdata_t *logdata, boolean gmt)
     ASSERT(logdata != NULL);
     ASSERT(logdata->raw_event != NULL);
     
+    if ( (nx_logdata_get_field_value(logdata, "Message", &msg) == TRUE) &&
+	 (msg.type == NX_VALUE_TYPE_STRING) && (msg.defined == TRUE) )
+    {
+	// we have a Message field
+    }
+    else
+    { // otherwise use raw_event
+	tmpmsg = nx_string_clone(logdata->raw_event);
+	msg.string = tmpmsg;
+	msg.type = NX_VALUE_TYPE_STRING;
+    }
+
     pri = nx_syslog_get_priority(logdata);
 
     if ( (nx_logdata_get_field_value(logdata, "EventTime", &timestamp) == TRUE) &&
-	 (timestamp.type = NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
+	 (timestamp.type == NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
     {
 	nx_date_to_rfc5424(tmpstr, sizeof(tmpstr), gmt, timestamp.datetime);
     }
@@ -1281,25 +1320,28 @@ void nx_logdata_to_syslog_rfc5424(nx_logdata_t *logdata, boolean gmt)
 
     nx_syslog_add_structured_data(logdata);
 
-    if ( (nx_logdata_get_field_value(logdata, "Message", &message) == TRUE) &&
-	 (message.type == NX_VALUE_TYPE_STRING) && (message.defined == TRUE) )
-    {
-	i = (int) logdata->raw_event->len;
-	nx_string_append(logdata->raw_event, " ", 1);
-	nx_string_append(logdata->raw_event, message.string->buf, (int) message.string->len);
-	for ( ; i < (int) logdata->raw_event->len; i++ )
-	{ // replace linebreaks with space
-	    if ( (logdata->raw_event->buf[i] == '\n') || (logdata->raw_event->buf[i] == '\r') )
-	    {
-		logdata->raw_event->buf[i] = ' ';
-	    }
+    // Append message
+    i = (int) logdata->raw_event->len;
+    nx_string_append(logdata->raw_event, " ", 1);
+    nx_string_append(logdata->raw_event, msg.string->buf, (int) msg.string->len);
+    for ( ; i < (int) logdata->raw_event->len; i++ )
+    { // replace linebreaks with space
+	if ( (logdata->raw_event->buf[i] == '\n') || (logdata->raw_event->buf[i] == '\r') )
+	{
+	    logdata->raw_event->buf[i] = ' ';
 	}
     }
+
+    if (tmpmsg != NULL)
+    { // clean up temp copy
+	nx_string_free(tmpmsg);
+    }
 }
 
 
 
 void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
+				uint64_t evtcnt,
 				char delimiter,
 				char replacement)
 {
@@ -1318,7 +1360,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
     pri = nx_syslog_get_priority(logdata);
 
     if ( (nx_logdata_get_field_value(logdata, "EventTime", &timestamp) == TRUE) &&
-	 (timestamp.type = NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
+	 (timestamp.type == NX_VALUE_TYPE_DATETIME) && (timestamp.defined == TRUE) )
     {
 	eventtime = timestamp.datetime;
 	nx_date_to_rfc3164(tmpstr, sizeof(tmpstr), eventtime);
@@ -1352,9 +1394,10 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 3. Criticality
     if ( (nx_logdata_get_field_value(logdata, "SeverityValue", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_INTEGER) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_INTEGER) && (tmpval.defined == TRUE) )
     {
-	apr_snprintf(tmpstr, sizeof(tmpstr), "%d", (int) tmpval.integer);
+	// SeverityValue starts from 1 and Criticality is from 0
+	apr_snprintf(tmpstr, sizeof(tmpstr), "%d", (int) tmpval.integer - 1); 
 	nx_string_append(logdata->raw_event, tmpstr, -1);
     }
     else
@@ -1366,9 +1409,14 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
     i = (int) logdata->raw_event->len;
 
     // 4. SourceName
-    if ( (nx_logdata_get_field_value(logdata, "FileName", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
-    {
+    if ( (nx_logdata_get_field_value(logdata, "Channel", &tmpval) == TRUE) &&
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+    { // This one is for im_msvistalog
+	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
+    }
+    else if ( (nx_logdata_get_field_value(logdata, "FileName", &tmpval) == TRUE) &&
+	      (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+    { // This one is for im_mseventlog
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
     else
@@ -1378,18 +1426,19 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
     nx_string_append(logdata->raw_event, delimiterstr, 1);
 
     // 5. Snare Event Counter
-    // we could use module->evt_recvd here, but is complicated so this is set to N/A
-    nx_string_append(logdata->raw_event, "N/A", -1);
-     nx_string_append(logdata->raw_event, delimiterstr, 1);
+    // use module->evt_recvd 
+    apr_snprintf(tmpstr, sizeof(tmpstr), "%lu", evtcnt); 
+    nx_string_append(logdata->raw_event, tmpstr, -1);
+    nx_string_append(logdata->raw_event, delimiterstr, 1);
 
     // 6. DateTime
-    nx_date_to_rfc3164_wday_year(tmpstr, sizeof(tmpstr), eventtime);
+    nx_date_to_rfc3164_wday_year(tmpstr, sizeof(tmpstr), eventtime, FALSE);
     nx_string_append(logdata->raw_event, tmpstr, -1);
     nx_string_append(logdata->raw_event, delimiterstr, 1);
 
     // 7. EventID
     if ( (nx_logdata_get_field_value(logdata, "EventID", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_INTEGER) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_INTEGER) && (tmpval.defined == TRUE) )
     {
 	apr_snprintf(tmpstr, sizeof(tmpstr), "%d", (int) tmpval.integer);
 	nx_string_append(logdata->raw_event, tmpstr, -1);
@@ -1402,7 +1451,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 8. SourceName
     if ( (nx_logdata_get_field_value(logdata, "SourceName", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
@@ -1414,7 +1463,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 9. UserName
     if ( (nx_logdata_get_field_value(logdata, "AccountName", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
@@ -1426,7 +1475,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 10. SIDType
     if ( (nx_logdata_get_field_value(logdata, "AccountType", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
@@ -1438,9 +1487,40 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 11. EventLogType
     if ( (nx_logdata_get_field_value(logdata, "EventType", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
-	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
+	if ( strcmp(tmpval.string->buf, "AUDIT_SUCCESS") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Success Audit", 13);
+	}
+	else if ( strcmp(tmpval.string->buf, "AUDIT_FAILURE") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Failure Audit", 13);
+	}
+	else if ( strcmp(tmpval.string->buf, "INFO") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Information", 11);
+	}
+	else if ( strcmp(tmpval.string->buf, "CRITICAL") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Critical", 8);
+	}
+	else if ( strcmp(tmpval.string->buf, "ERROR") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Error", 5);
+	}
+	else if ( strcmp(tmpval.string->buf, "WARNING") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Warning", 7);
+	}
+	else if ( strcmp(tmpval.string->buf, "VERBOSE") == 0 )
+	{
+	    nx_string_append(logdata->raw_event, "Verbose", 7);
+	}
+	else
+	{
+	    nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
+	}
     }
     else
     {
@@ -1450,7 +1530,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 12. ComputerName
     if ( (nx_logdata_get_field_value(logdata, "Hostname", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
@@ -1462,7 +1542,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 13. CategoryString
     if ( (nx_logdata_get_field_value(logdata, "Category", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	nx_string_append(logdata->raw_event, tmpval.string->buf, (int) tmpval.string->len);
     }
@@ -1477,7 +1557,7 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 
     // 15. ExpandedString
     if ( (nx_logdata_get_field_value(logdata, "Message", &tmpval) == TRUE) &&
-	 (tmpval.type = NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
+	 (tmpval.type == NX_VALUE_TYPE_STRING) && (tmpval.defined == TRUE) )
     {
 	for ( i = 0; i < (int) tmpval.string->len; i++ )
 	{
@@ -1493,8 +1573,19 @@ void nx_logdata_to_syslog_snare(nx_logdata_t *logdata,
 	nx_string_append(logdata->raw_event, "N/A", -1);
     }
     nx_string_append(logdata->raw_event, delimiterstr, 1);
+
     // Eventlog Counter
-    nx_string_append(logdata->raw_event, "N/A", -1);
+    if ( (nx_logdata_get_field_value(logdata, "RecordNumber", &tmpval) == TRUE) &&
+	 (tmpval.type == NX_VALUE_TYPE_INTEGER) && (tmpval.defined == TRUE) )
+    {
+	apr_snprintf(tmpstr, sizeof(tmpstr), "%d", (int) tmpval.integer);
+	nx_string_append(logdata->raw_event, tmpstr, -1);
+    }
+    else
+    {
+	nx_string_append(logdata->raw_event, "N/A", -1);
+    }
+
 
     for ( i = 0; i < (int) logdata->raw_event->len; i++ )
     { // replace linebreaks with space
