@@ -22,11 +22,14 @@ typedef struct nx_im_dbi_option_t
 
 typedef struct nx_im_dbi_conf_t
 {
+#ifdef HAVE_DBI_INITIALIZE_R
+    dbi_inst dbi_inst;
+#endif
     const char *driver;
     apr_array_header_t *options;
     boolean savepos;
     int64_t last_id;
-    int	read_interval;
+    float poll_interval;
     const char *sql;
     char *_sql;
     size_t _sql_bufsize;
