@@ -143,6 +143,8 @@ static void im_ssl_accept(nx_module_t *module)
 			 "couldn't set SO_NONBLOCK on accepted socket");
 	    CHECKERR_MSG(apr_socket_timeout_set(sock, 0),
 			 "couldn't set socket timeout on accepted socket");
+	    CHECKERR_MSG(apr_socket_opt_set(sock, APR_SO_KEEPALIVE, 1),
+			 "couldn't set TCP_KEEPALIVE on accepted socket");
 	    CHECKERR_MSG(apr_socket_addr_get(&sa, APR_REMOTE, sock),
 			 "couldn't get info on accepted socket");
 	    CHECKERR_MSG(apr_sockaddr_ip_get(&ipstr, sa),
