@@ -93,12 +93,22 @@ static void om_http_disconnect(nx_module_t *module, boolean reconnect)
 
     modconf = (nx_om_http_conf_t *) module->config;
 
+    log_debug("om_http_disconnect");
+
     modconf->connected = FALSE;
     if ( modconf->sock != NULL )
     {
+	nx_module_pollset_remove_socket(module, modconf->sock);
 	apr_socket_close(modconf->sock);
 	modconf->sock = NULL;
     }
+
+    if ( modconf->ssl != NULL )
+    {
+	nx_ssl_destroy(&(modconf->ssl));
+	modconf->ssl = NULL;
+    }
+
     apr_pool_clear(modconf->pool);
 
     if ( reconnect == TRUE )
@@ -120,6 +130,8 @@ static void om_http_reset(nx_module_t *module)
 
     modconf = (nx_om_http_conf_t *) module->config;
 
+    modconf->logdata = NULL;
+
     if ( modconf->bio_resp_head != NULL )
     {
 	BIO_free_all(modconf->bio_resp_head);
@@ -215,6 +227,7 @@ static void om_http_send_request(nx_module_t *module)
     nx_om_http_conf_t *modconf;
     int sslrv;
     int nbytes2;
+    nx_exception_t e;
 
     modconf = (nx_om_http_conf_t *) module->config;
     
@@ -238,7 +251,7 @@ static void om_http_send_request(nx_module_t *module)
 		    nx_module_add_poll_event(module);
 		    break;
 		case SSL_ERROR_ZERO_RETURN: // disconnected
-		    log_warn("http server disconnected");
+		    log_warn("https server disconnected while sending the request");
 		    om_http_disconnect(module, TRUE);
 		    om_http_reset(module);
 		    break;
@@ -251,9 +264,18 @@ static void om_http_send_request(nx_module_t *module)
 		    nx_module_add_poll_event(module);
 		    break;
 		default:
+		    try
+		    {
+			nx_ssl_check_io_error(modconf->ssl, sslrv);
+		    }
+		    catch(e)
+		    {
+			om_http_disconnect(module, TRUE);
+			om_http_reset(module);
+			rethrow(e);
+		    }
 		    om_http_disconnect(module, TRUE);
 		    om_http_reset(module);
-		    nx_ssl_check_io_error(modconf->ssl, sslrv);
 		    break;
 	    }
 	}
@@ -318,6 +340,12 @@ static void om_http_stop(nx_module_t *module)
 	    modconf->sock = NULL;
 	}
     }
+    if ( modconf->ssl != NULL )
+    {
+	nx_ssl_destroy(&(modconf->ssl));
+	modconf->ssl = NULL;
+    }
+
     modconf->connected = FALSE;
     apr_pool_clear(modconf->pool);
 }
@@ -330,29 +358,10 @@ static void io_err_handler(nx_module_t *module, nx_exception_t *e)
     ASSERT(e != NULL);
     ASSERT(module != NULL);
 
-    if ( APR_STATUS_IS_ECONNREFUSED(e->code) ||
-	 APR_STATUS_IS_ECONNABORTED(e->code) ||
-	 APR_STATUS_IS_ECONNRESET(e->code) ||
-	 APR_STATUS_IS_ETIMEDOUT(e->code) ||
-	 APR_STATUS_IS_TIMEUP(e->code) ||
-	 APR_STATUS_IS_EHOSTUNREACH(e->code) ||
-	 APR_STATUS_IS_ENETUNREACH(e->code) ||
-	 APR_STATUS_IS_EOF(e->code) ||
-	 APR_STATUS_IS_EPIPE(e->code) ||
-	 (e->code == APR_SUCCESS) )
-    {
-	om_http_disconnect(module, TRUE);
-	rethrow(*e);
-    }
-    if ( APR_STATUS_IS_EAGAIN(e->code) )
-    {
-	nx_panic("got EAGAIN, bug in non-blocking code");
-    }
-    //default:
     nx_module_stop_self(module);
     om_http_stop(module);
-    rethrow_msg(*e, "fatal connection error, reconnection will not be attempted (statuscode: %d)",
-		e->code);
+    om_http_add_reconnect_event(module);
+    rethrow(*e);
 }
 
 
@@ -394,6 +403,7 @@ static void om_http_connect(nx_module_t *module)
 
 	if ( modconf->server.https == TRUE )
 	{
+	    ASSERT(modconf->ssl == NULL);
 	    modconf->ssl = nx_ssl_from_socket(&(modconf->ssl_ctx), modconf->sock);
 	    CHECKERR_MSG(apr_socket_opt_set(modconf->sock, APR_SO_NONBLOCK, 0),
 			 "couldn't set SO_NONBLOCK on connecting socket");
@@ -413,6 +423,8 @@ static void om_http_connect(nx_module_t *module)
 		     "couldn't set SO_NONBLOCK on tcp socket");
 	CHECKERR_MSG(apr_socket_timeout_set(modconf->sock, 0),
 		     "couldn't set socket timeout on tcp socket");
+	CHECKERR_MSG(apr_socket_opt_set(modconf->sock, APR_SO_KEEPALIVE, 1),
+		     "couldn't set TCP_KEEPALIVE on connecting socket");
 	modconf->reconnect = 0;
 
 	// POLLIN is to detect disconnection
@@ -483,8 +495,6 @@ static void om_http_data_available(nx_module_t *module)
 	om_http_send_request(module);
     }
 
-    log_debug("request sent");
-
     BIO_free_all(modconf->bio_req);
     modconf->bio_req = NULL;
 }
@@ -570,11 +580,17 @@ static void om_http_read_response(nx_module_t *module)
     apr_size_t len, i;
     int nbytes;
     int sslrv;
+    nx_exception_t e;
 
     log_debug("nx_om_http_read_response()");
     
     modconf = (nx_om_http_conf_t *) module->config;
 
+    if ( modconf->connected == FALSE )
+    {
+	return;
+    }
+
     len = NX_OM_HTTP_RESPBUFSIZE;
     memset(modconf->respbuf, 0, len);
 
@@ -587,7 +603,7 @@ static void om_http_read_response(nx_module_t *module)
 	    case SSL_ERROR_NONE:
 		break;
 	    case SSL_ERROR_ZERO_RETURN: // disconnected
-		log_warn("http server disconnected");
+		log_warn("https server disconnected while reading the response");
 		om_http_disconnect(module, TRUE);
 		if ( nbytes == 0 )
 		{
@@ -600,9 +616,18 @@ static void om_http_read_response(nx_module_t *module)
 		log_debug("got EAGAIN for nonblocking read in module %s", module->name);
 		break;
 	    default:
+		try
+		{
+		    nx_ssl_check_io_error(modconf->ssl, sslrv);
+		}
+		catch(e)
+		{
+		    om_http_disconnect(module, TRUE);
+		    om_http_reset(module);
+		    rethrow(e);
+		}
 		om_http_disconnect(module, TRUE);
 		om_http_reset(module);
-		nx_ssl_check_io_error(modconf->ssl, sslrv);
 		return;
 	}
 	len = (apr_size_t) nbytes;
@@ -629,7 +654,7 @@ static void om_http_read_response(nx_module_t *module)
 
 	if ( got_eof == TRUE )
 	{
-	    log_warn("http server disconnected");
+	    log_warn("http server disconnected while reading the response");
 	    om_http_disconnect(module, TRUE);
 	    if ( len == 0 )
 	    {
@@ -737,13 +762,19 @@ static void om_http_read_response(nx_module_t *module)
 	}
     }
 
-    log_debug("parsed resp");
-    modconf->resp_wait = FALSE;
+    if ( (modconf->got_resp_body == TRUE) && (modconf->got_resp_head == TRUE) )
+    {
+	log_debug("om_http received and parsed response");
 
-    nx_module_logqueue_pop(module, modconf->logdata);
-    nx_logdata_free(modconf->logdata);
-    modconf->logdata = NULL;
-    om_http_reset(module);
+	modconf->resp_wait = FALSE;
+
+	if (modconf->logdata != NULL)
+	{
+	    nx_module_logqueue_pop(module, modconf->logdata);
+	    nx_logdata_free(modconf->logdata);
+	}
+	om_http_reset(module);
+    }
     nx_module_data_available(module);
     nx_module_add_poll_event(module);
 }
@@ -1029,12 +1060,17 @@ static void om_http_event(nx_module_t *module, nx_event_t *event)
 	    om_http_send_request(module);
 	    break;
 	case NX_EVENT_RECONNECT:
-	    om_http_connect(module);
+	    nx_module_start_self(module);
+	    modconf->reconnect = 0;
+	    om_http_data_available(module);
 	    break;
 	case NX_EVENT_DISCONNECT:
-	    om_http_disconnect(module, TRUE);
-	    om_http_reset(module);
-	    log_error("disconnected");
+	    if ( modconf->connected == TRUE )
+	    { // the disconnection may have been already noticed
+		om_http_disconnect(module, TRUE);
+		om_http_reset(module);
+		log_error("http server disconnected");
+	    }
 	    break;
 	case NX_EVENT_TIMEOUT:
 	    modconf->timeout_event = NULL;
