@@ -17,6 +17,11 @@ typedef struct nx_om_tcp_conf_t
     apr_socket_t	*sock;
     boolean		connected;
     int			reconnect;
+
+    boolean		listen; ///< listen for connections (a.k.a. pull-mode)
+    boolean		queueinlistenmode; ///< do not drop messages in listen mode
+    apr_socket_t	*listensock;
+    nx_module_input_list_t *connections; 
 } nx_om_tcp_conf_t;
 
 
