@@ -118,6 +118,7 @@ static apr_size_t pm_buffer_pop_disk(nx_module_t *module, nx_logdata_t **logdata
 
     modconf = (nx_pm_buffer_conf_t *) module->config;
 
+    *logdata = NULL;
     if ( modconf->pop_file == NULL )
     {
 	ASSERT(modconf->push_id > 0);
@@ -125,7 +126,6 @@ static apr_size_t pm_buffer_pop_disk(nx_module_t *module, nx_logdata_t **logdata
 	if ( modconf->pop_id > modconf->push_id )
 	{
 	    // nothing to read
-	    *logdata = NULL;
 	    return ( 0 );
 	}
 	    
@@ -148,7 +148,6 @@ static apr_size_t pm_buffer_pop_disk(nx_module_t *module, nx_logdata_t **logdata
 	if ( modconf->pop_count == modconf->push_count )
 	{
 	    // nothing to read
-	    *logdata = NULL;
 	    return ( 0 );
 	}
     }
@@ -229,7 +228,7 @@ static apr_size_t pm_buffer_pop_disk(nx_module_t *module, nx_logdata_t **logdata
 
 static void pm_buffer_data_available(nx_module_t *module)
 {
-    nx_logdata_t *logdata = NULL;
+    nx_logdata_t *logdata;
     nx_pm_buffer_conf_t *modconf;
     boolean can_send;
     nx_module_status_t status;
@@ -265,11 +264,12 @@ static void pm_buffer_data_available(nx_module_t *module)
 	else
 	{ // DISK
 	    bytesread = pm_buffer_pop_disk(module, &logdata);
-	
+
 	    if ( logdata == NULL )
 	    {
 		break;
 	    }
+
 	    ASSERT(modconf->buffer_size >= bytesread);
 	    modconf->buffer_size -= bytesread;
 	    nx_module_progress_logdata(module, logdata);
@@ -327,6 +327,7 @@ static void pm_buffer_data_available(nx_module_t *module)
 	    
 	    written = pm_buffer_push_disk(module, logdata);
 	    nx_module_logqueue_pop(module, logdata);
+	    nx_logdata_free(logdata);
 	    modconf->buffer_size += written;
 	}
 
@@ -479,7 +480,7 @@ static void pm_buffer_start(nx_module_t *module)
 
     if ( modconf->type == NX_PM_BUFFER_TYPE_DISK )
     {
-	pool = nx_pool_create_child(module->pool);
+	pool = nx_pool_create_core();
 
 	try
 	{
@@ -490,7 +491,7 @@ static void pm_buffer_start(nx_module_t *module)
 	    apr_off_t offs = 0;
 
 	    CHECKERR_MSG(apr_dir_open(&dir, modconf->basedir, pool),
-			 "failed to read disk buffer files");
+			 "failed to read disk buffer files from '%s'", modconf->basedir);
 	
 	    namelen = (size_t) apr_snprintf(fname, sizeof(fname), "%s.", module->name);
 
@@ -576,8 +577,7 @@ static void pm_buffer_start(nx_module_t *module)
 	    {
 		throw_msg("disk buffer filename length limit exceeeded");
 	    }
-	    pool = nx_pool_create_child(module->pool);
-	    
+	    pool = nx_pool_create_core();
 	    CHECKERR_MSG(apr_file_open(&(modconf->push_file), filename, APR_WRITE | APR_CREATE | APR_TRUNCATE,
 				       APR_OS_DEFAULT, pool),
 			 "couldn't open disk buffer file %s for writing", filename);
@@ -643,6 +643,26 @@ static void pm_buffer_stop(nx_module_t *module)
 		apr_pool_destroy(pool);
 	    }
 	}
+	if ( modconf->pop_file != NULL )
+	{
+	    pool = apr_file_pool_get(modconf->pop_file);
+	    apr_file_close(modconf->pop_file);
+	    modconf->pop_file = NULL;
+	    apr_pool_destroy(pool);
+	}
+
+	if ( modconf->push_file != NULL )
+	{
+	    pool = apr_file_pool_get(modconf->push_file);
+	    apr_file_close(modconf->push_file);
+	    modconf->push_file = NULL;
+	    apr_pool_destroy(pool);
+	}
+
+	modconf->pop_id = 0;
+	modconf->push_id = 0;
+	modconf->pop_count = 0;
+	modconf->push_count = 0;
     }
     else if ( modconf->type == NX_PM_BUFFER_TYPE_MEM )
     {
