@@ -602,6 +602,10 @@ static void parse_condition(nx_module_t *module,
     {
 	*cond = nx_expr_parse(module, curr->args, pool, curr->filename,
 			      curr->line_num, curr->argsstart);
+	if ( *cond == NULL )
+	{
+	    throw_msg("invalid or empty expression for Condition: '%s'", curr->args);
+	}
 	if ( (*cond)->rettype != NX_VALUE_TYPE_BOOLEAN )
 	{
 	    throw_msg("boolean type required in expression, found '%s'",
@@ -700,6 +704,10 @@ static void pm_evcorr_config(nx_module_t *module)
 		    rule->suppressed.context_expr = nx_expr_parse(module, curr2->args, module->pool,
 								  curr2->filename, curr2->line_num,
 								  curr2->argsstart);
+		    if ( rule->suppressed.context_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for Context: '%s'", curr2->args);
+		    }
 		    if ( rule->suppressed.context_expr->type == NX_EXPR_TYPE_VALUE )
 		    {
 			nx_conf_error(curr2, "using a constant value for 'Context' is useless");
@@ -760,6 +768,10 @@ static void pm_evcorr_config(nx_module_t *module)
 		    rule->pair.context_expr = nx_expr_parse(module, curr2->args, module->pool,
 							    curr2->filename, curr2->line_num,
 							    curr2->argsstart);
+		    if ( rule->pair.context_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for Context: '%s'", curr2->args);
+		    }
 		    if ( rule->pair.context_expr->type == NX_EXPR_TYPE_VALUE )
 		    {
 			nx_conf_error(curr2, "using a constant value for 'Context' is useless");
@@ -828,6 +840,10 @@ static void pm_evcorr_config(nx_module_t *module)
 		    rule->absence.context_expr = nx_expr_parse(module, curr2->args, module->pool,
 							       curr2->filename, curr2->line_num,
 							       curr2->argsstart);
+		    if ( rule->absence.context_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for Context: '%s'", curr2->args);
+		    }
 		    if ( rule->absence.context_expr->type == NX_EXPR_TYPE_VALUE )
 		    {
 			nx_conf_error(curr2, "using a constant value for 'Context' is useless");
@@ -901,6 +917,10 @@ static void pm_evcorr_config(nx_module_t *module)
 		    rule->thresholded.context_expr = nx_expr_parse(module, curr2->args, module->pool,
 								   curr2->filename, curr2->line_num,
 								   curr2->argsstart);
+		    if ( rule->thresholded.context_expr == NULL )
+		    {
+			throw_msg("invalid or empty expression for Context: '%s'", curr2->args);
+		    }
 		    if ( rule->thresholded.context_expr->type == NX_EXPR_TYPE_VALUE )
 		    {
 			nx_conf_error(curr2, "using a constant value for 'Context' is useless");
@@ -1098,6 +1118,7 @@ static void pm_evcorr_absence_clean(nx_module_t *module,
 	    {
 		nx_event_remove(absence->event);
 		nx_event_free(absence->event);
+		absence->event = NULL;
 	    }
 	    free_hash_item(rule->absence.context, val, key, keylen);
 	}
@@ -1117,6 +1138,7 @@ static void pm_evcorr_absence_clean(nx_module_t *module,
 		{
 		    nx_event_remove(absence->event);
 		    nx_event_free(absence->event);
+		    absence->event = NULL;
 		}
 		free_hash_item(rule->absence.context, val, key, keylen);
 	    }
@@ -1147,6 +1169,7 @@ static void pm_evcorr_thresholded_clean(nx_module_t *module,
 	    if ( thresholded->matchlist != NULL )
 	    {
 		free(thresholded->matchlist);
+		thresholded->matchlist = NULL;
 	    }
 	    free_hash_item(rule->thresholded.context, val, key, keylen);
 	}
@@ -1160,12 +1183,15 @@ static void pm_evcorr_thresholded_clean(nx_module_t *module,
 	    idx = apr_hash_next(idx);
 	    modconf = (nx_pm_evcorr_conf_t *) module->config;
 
-	    if ( thresholded->matchlist[thresholded->num_match - 1] + 
-		 (rule->thresholded.interval * APR_USEC_PER_SEC) <= modconf->timeval )
+	    ASSERT(thresholded != NULL);
+	    if ( (thresholded->matchlist != NULL) &&
+		 ((thresholded->matchlist[thresholded->num_match - 1] + 
+		   (rule->thresholded.interval * APR_USEC_PER_SEC)) <= modconf->timeval) )
 	    {
 		if ( thresholded->matchlist != NULL )
 		{
 		    free(thresholded->matchlist);
+		    thresholded->matchlist = NULL;
 		}
 		free_hash_item(rule->thresholded.context, val, key, keylen);
 	    }
@@ -1284,7 +1310,8 @@ static void pm_evcorr_stop(nx_module_t *module)
 
     if ( modconf->cleanup_event != NULL )
     {
-	// event already removed by core
+	nx_event_remove(modconf->cleanup_event);
+	nx_event_free(modconf->cleanup_event);
 	modconf->cleanup_event = NULL;
     }
 
