@@ -113,7 +113,7 @@ static void nx_pattern_matchfield_compile(apr_pool_t *pool,
 {
     const char *error = NULL;
     int erroroffs = 0;
-    int size;
+    size_t size;
     pcre *tmppcre;
     int rc;
     int capturecount;
@@ -202,11 +202,14 @@ void nx_pattern_field_list_insert_field(apr_pool_t *pool,
     ASSERT(value != NULL);
     
     val = nx_value_from_string(value, type); //FIXME: use pool alloc here
-    field = apr_pcalloc(pool, sizeof(nx_logdata_field_t));
-    field->value = val;
-    field->key = apr_pstrdup(pool, key);
+    if ( val != NULL )
+    {
+	field = apr_pcalloc(pool, sizeof(nx_logdata_field_t));
+	field->value = val;
+	field->key = apr_pstrdup(pool, key);
 
-    NX_DLIST_INSERT_TAIL(fields, field, link);
+	NX_DLIST_INSERT_TAIL(fields, field, link);
+    }
 }
 
 
